########################################################################
# File name: test_form.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import copy
import unittest
import unittest.mock

import aioxmpp

import aioxmpp.forms.fields as fields
import aioxmpp.forms.form as form
import aioxmpp.forms.xso as forms_xso


class FakeDescriptor(fields.AbstractDescriptor):
    def __init__(self, key):
        self._keys = [key]

    def descriptor_keys(self):
        return self._keys


class TestDescriptorClass(unittest.TestCase):
    def test_is_abcmeta(self):
        self.assertTrue(issubclass(
            form.DescriptorClass,
            abc.ABCMeta,
        ))

    def test_init(self):
        class Cls(metaclass=form.DescriptorClass):
            pass

        self.assertDictEqual(
            Cls.DESCRIPTOR_MAP,
            {},
        )

        self.assertSequenceEqual(
            list(Cls.DESCRIPTORS),
            [],
        )

        self.assertSequenceEqual(
            Cls.__slots__,
            (),
        )

    def test_use_slots_from_declaration(self):
        class Cls(metaclass=form.DescriptorClass):
            __slots__ = ("foo", )

        self.assertSequenceEqual(
            Cls.__slots__,
            ("foo",)
        )

    def test_inherit_slots(self):
        class Base(metaclass=form.DescriptorClass):
            __slots__ = ("foo", )

        class Child(Base):
            pass

        self.assertSequenceEqual(
            Child.__slots__,
            ()
        )

    def test_disable_slots_via_kwarg(self):
        class Cls(metaclass=form.DescriptorClass, protect=False):
            pass

        self.assertFalse(hasattr(Cls, "__slots__"))

    def test_collect_descriptors(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor(("foo", "bar"))
            x2 = FakeDescriptor(("baz",))

        self.assertDictEqual(
            Cls.DESCRIPTOR_MAP,
            {
                ("foo", "bar"): Cls.x1,
                ("baz", ): Cls.x2,
            }
        )

        self.assertSetEqual(
            set(Cls.DESCRIPTORS),
            {
                Cls.x1,
                Cls.x2,
            }
        )

    def test_single_inheritance(self):
        class ClsA(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("x1")

        class ClsB(ClsA):
            x2 = FakeDescriptor("x2")

        self.assertDictEqual(
            ClsA.DESCRIPTOR_MAP,
            {
                "x1": ClsA.x1,
            }
        )

        self.assertSetEqual(
            ClsA.DESCRIPTORS,
            {
                ClsA.x1,
            }
        )

        self.assertDictEqual(
            ClsB.DESCRIPTOR_MAP,
            {
                "x1": ClsA.x1,
                "x2": ClsB.x2,
            }
        )

        self.assertSetEqual(
            ClsB.DESCRIPTORS,
            {
                ClsA.x1,
                ClsB.x2,
            }
        )

    def test_multi_inheritance(self):
        class ClsA(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("x1")

        class ClsB(ClsA):
            x2 = FakeDescriptor("x2")

        class ClsC(ClsA):
            x3 = FakeDescriptor("x3")

        class ClsD(ClsB, ClsC):
            x4 = FakeDescriptor("x4")

        self.assertDictEqual(
            ClsD.DESCRIPTOR_MAP,
            {
                "x1": ClsD.x1,
                "x2": ClsD.x2,
                "x3": ClsD.x3,
                "x4": ClsD.x4,
            }
        )

        self.assertSetEqual(
            ClsD.DESCRIPTORS,
            {
                ClsD.x1,
                ClsD.x2,
                ClsD.x3,
                ClsD.x4,
            }
        )

    def test_reject_conflicts(self):
        with self.assertRaisesRegex(
                TypeError,
                "descriptor with key .* already declared at .*"):
            class ClsA(metaclass=form.DescriptorClass):
                x1 = FakeDescriptor("x1")
                x2 = FakeDescriptor("x1")

    def test_reject_conflicts_at_multi_inheritance(self):
        class ClsA(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("x1")

        class ClsB(metaclass=form.DescriptorClass):
            x2 = FakeDescriptor("x1")

        with self.assertRaisesRegex(
                TypeError,
                "descriptor with key .* already declared at .*"):
            class ClsC(ClsA, ClsB):
                pass

    def test_late_addition_of_descriptors(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("foo")

        Cls.x2 = FakeDescriptor("bar")

        self.assertDictEqual(
            Cls.DESCRIPTOR_MAP,
            {
                "foo": Cls.x1,
                "bar": Cls.x2,
            }
        )

        self.assertSetEqual(
            Cls.DESCRIPTORS,
            {
                Cls.x1,
                Cls.x2,
            }
        )

    def test_reject_conflict_on_late_addition_of_descriptors(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("foo")

        with self.assertRaisesRegex(
                TypeError,
                r"descriptor with key .* already declared at .*"):
            Cls.x2 = FakeDescriptor("foo")

        self.assertDictEqual(
            Cls.DESCRIPTOR_MAP,
            {
                "foo": Cls.x1,
            }
        )

        self.assertSetEqual(
            Cls.DESCRIPTORS,
            {
                Cls.x1,
            }
        )

    def test_reject_late_addition_for_subclassed(self):
        class ClsA(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("foo")

        class ClsB(ClsA):
            pass

        with self.assertRaisesRegex(
                TypeError,
                r"cannot add descriptors to classes with subclasses"):
            ClsA.x2 = FakeDescriptor("bar")

    def test_reject_removal_of_descriptors(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("x1")

        with self.assertRaisesRegex(
                AttributeError,
                r"removal of descriptors is not allowed"):
            del Cls.x1

    def test_allow_removal_of_other_attributes(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("x1")
            x2 = "foo"

        del Cls.x2
        self.assertFalse(hasattr(Cls, "x2"))

    def test_reject_overwriting_descriptor_attribute(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("x1")

        with self.assertRaisesRegex(
                AttributeError,
                r"descriptor attributes cannot be set"):
            Cls.x1 = "foo"

    def test_allow_setting_and_overwriting_other_attributes(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("x1")

        Cls.x2 = "foo"
        Cls.x3 = "bar"
        Cls.x2 = 10

    def test_set_attribute_name_on_descriptors(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("foo")
            x2 = FakeDescriptor("bar")

        self.assertEqual(Cls.x1.attribute_name, "x1")
        self.assertEqual(Cls.x2.attribute_name, "x2")

    def test_set_root_class_on_descriptors(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("foo")
            x2 = FakeDescriptor("bar")

        self.assertIs(Cls.x1.root_class, Cls)
        self.assertIs(Cls.x2.root_class, Cls)

    def test_reject_descriptor_used_twice(self):
        d = FakeDescriptor("foo")

        class ClsA(metaclass=form.DescriptorClass):
            x1 = d

        with self.assertRaisesRegex(
                ValueError,
                r"descriptor cannot be used on multiple classes"):
            class ClsB(metaclass=form.DescriptorClass):
                x1 = d

    def test_register_descriptor_keys(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("foo")

        Cls._register_descriptor_keys(
            Cls.x1,
            ["bar", "baz"]
        )

        self.assertDictEqual(
            Cls.DESCRIPTOR_MAP,
            {
                "foo": Cls.x1,
                "bar": Cls.x1,
                "baz": Cls.x1,
            }
        )

    def test_reject_conflict_on__register_descriptor_keys(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("foo")
            x2 = FakeDescriptor("bar")

        with self.assertRaisesRegex(
                TypeError,
                r"descriptor with key .* already declared at .*"):
            Cls._register_descriptor_keys(Cls.x1, ["bar"])

    def test_reject__register_descriptor_keys_for_subclassed(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("foo")

        class Subclass(Cls):
            pass

        with self.assertRaisesRegex(
                TypeError,
                r"descriptors cannot be modified on classes with subclasses"):
            Cls._register_descriptor_keys(Cls.x1, ["bar"])

    def test_reject__register_descriptor_keys_if_not_root_class(self):
        class Cls(metaclass=form.DescriptorClass):
            x1 = FakeDescriptor("foo")

        class Subclass(Cls):
            pass

        with self.assertRaisesRegex(
                TypeError,
                r"descriptors cannot be modified on classes with subclasses"):
            Subclass._register_descriptor_keys(
                Subclass.x1,
                ["bar"]
            )


class TestFormClass(unittest.TestCase):
    def test_is_DescriptorClass(self):
        self.assertTrue(issubclass(
            form.FormClass,
            form.DescriptorClass,
        ))


class TestForm(unittest.TestCase):
    def test_init(self):
        class F(form.Form):
            field = fields.TextSingle(
                "foovar",
            )

        f = F()
        self.assertDictEqual(
            f._descriptor_data,
            {},
        )

        f2 = F()
        self.assertIsNot(
            f._descriptor_data,
            f2._descriptor_data
        )

        self.assertIsNone(
            f._recv_xso,
        )

    def test___new___creates_data_dict(self):
        class F(form.Form):
            pass

        f = F.__new__(F)
        self.assertDictEqual(f._descriptor_data, {})

    def test_from_xso_checks_FORM_TYPE(self):
        class F(form.Form):
            FORM_TYPE = "foo"

        tree = forms_xso.Data(type_=forms_xso.DataType.FORM)
        tree.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.HIDDEN,
                values=["bar"],
                var="FORM_TYPE"
            )
        )

        with self.assertRaisesRegex(
                ValueError,
                "mismatching FORM_TYPE"):
            F.from_xso(tree)

    def test_form_xso_allows_all_data_types(self):
        allow = [
            forms_xso.DataType.CANCEL,
            forms_xso.DataType.FORM,
            forms_xso.DataType.SUBMIT,
            forms_xso.DataType.RESULT,
        ]

        class F(form.Form):
            pass

        for t in forms_xso.DataType:
            tree = forms_xso.Data(type_=t)
            F.from_xso(tree)

    def test_from_xso_single_field(self):
        class F(form.Form):
            field = fields.TextSingle(
                "foobar",
            )

        tree = forms_xso.Data(type_=forms_xso.DataType.FORM)
        tree.fields.append(
            forms_xso.Field(values=["value"], var="foobar")
        )

        f = F.from_xso(tree)
        self.assertIsInstance(f, F)
        self.assertEqual(
            f.field.value,
            "value",
        )

    def test_from_xso_handles_missing_type_(self):
        class F(form.Form):
            field = fields.ListSingle(
                "foobar",
            )

        tree = forms_xso.Data(type_=forms_xso.DataType.FORM)
        tree.fields.append(
            forms_xso.Field(type_=None,
                            values=["value"],
                            var="foobar")
        )

        f = F.from_xso(tree)
        self.assertIsInstance(f, F)
        self.assertEqual(
            f.field.value,
            "value",
        )

    def test_copies_have_independent_descriptor_data(self):
        class F(form.Form):
            field = fields.TextSingle(
                "foovar",
            )

        f = F()
        f.field.value = "foo"

        copied = copy.copy(f)
        self.assertDictEqual(
            f._descriptor_data,
            copied._descriptor_data,
        )

        self.assertIsNot(
            f._descriptor_data,
            copied._descriptor_data,
        )

    def test_deepcopy_does_not_copy_descriptors(self):
        class F(form.Form):
            field = fields.TextSingle(
                "foovar",
            )

        f = F()
        f.field.value = "foo"

        copied = copy.deepcopy(f)
        self.assertIsNot(
            f._descriptor_data,
            copied._descriptor_data,
        )

        self.assertIsNot(
            f.field,
            copied.field,
        )

    def test_from_xso_complex(self):
        data = forms_xso.Data(type_=forms_xso.DataType.FORM)

        data.fields.append(
            forms_xso.Field(
                var="FORM_TYPE",
                type_=forms_xso.FieldType.HIDDEN,
                values=["some-uri"],
            )
        )

        data.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.FIXED,
                values=["This is some heading."],
            )
        )

        data.fields.append(
            forms_xso.Field(
                var="jid",
                type_=forms_xso.FieldType.JID_SINGLE,
                values=[],
                desc="some description",
                label="some label",
            )
        )

        class F(form.Form):
            jid = fields.JIDSingle(
                var="jid",
            )

        f = F.from_xso(data)
        self.assertIsNone(f.jid.value)
        self.assertIs(f._recv_xso, data)

    def test_from_xso_rejects_mismatching_type(self):
        data = forms_xso.Data(type_=forms_xso.DataType.FORM)

        data.fields.append(
            forms_xso.Field(
                var="FORM_TYPE",
                type_=forms_xso.FieldType.HIDDEN,
                values=["some-uri"],
            )
        )

        data.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.FIXED,
                values=["This is some heading."],
            )
        )

        data.fields.append(
            forms_xso.Field(
                var="jid",
                type_=forms_xso.FieldType.JID_SINGLE,
                values=[],
                desc="some description",
                label="some label",
            )
        )

        class F(form.Form):
            jid = fields.TextSingle(
                var="jid",
            )

        with self.assertRaisesRegex(
                ValueError,
                r"mismatching type (.+ != .+) on field .+"):
            F.from_xso(data)

    def test_from_xso_allows_upcast(self):
        data = forms_xso.Data(type_=forms_xso.DataType.FORM)

        data.fields.append(
            forms_xso.Field(
                var="FORM_TYPE",
                type_=forms_xso.FieldType.HIDDEN,
                values=["some-uri"],
            )
        )

        data.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.FIXED,
                values=["This is some heading."],
            )
        )

        data.fields.append(
            forms_xso.Field(
                var="jid",
                type_=forms_xso.FieldType.TEXT_SINGLE,
                values=[],
                desc="some description",
                label="some label",
            )
        )

        class F(form.Form):
            jid = fields.TextPrivate(
                var="jid",
            )

        F.from_xso(data)

    def test_render_reply(self):
        data = forms_xso.Data(type_=forms_xso.DataType.FORM)

        data.fields.append(
            forms_xso.Field(
                var="FORM_TYPE",
                type_=forms_xso.FieldType.HIDDEN,
                values=["some-uri"],
            )
        )

        data.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.FIXED,
                values=["This is some heading."],
            )
        )

        data.fields.append(
            forms_xso.Field(
                var="jid",
                type_=forms_xso.FieldType.JID_SINGLE,
                values=[],
                desc="some description",
            )
        )

        class F(form.Form):
            jid = fields.JIDSingle(
                var="jid",
                label="Foobar"
            )

            other = fields.TextSingle(
                var="foo",
            )

        f = F.from_xso(data)
        f.jid.value = aioxmpp.JID.fromstr("foo@bar.baz")

        result = f.render_reply()
        self.assertIsInstance(result, forms_xso.Data)
        self.assertEqual(
            result.type_,
            forms_xso.DataType.SUBMIT,
        )
        self.assertEqual(
            len(result.fields),
            3
        )

        self.assertIs(
            data.fields[0],
            result.fields[0]
        )
        self.assertIs(
            data.fields[1],
            result.fields[1],
        )

        self.assertIsNot(
            data.fields[2],
            result.fields[2],
        )

        jid_field = result.fields[2]
        self.assertSequenceEqual(
            jid_field.values,
            ["foo@bar.baz"]
        )
        self.assertEqual(
            jid_field.desc,
            "some description",
        )
        self.assertIsNone(
            jid_field.label
        )

    def test_render_reply_includes_unknown_field(self):
        data = forms_xso.Data(type_=forms_xso.DataType.FORM)

        data.fields.append(
            forms_xso.Field(
                var="jid",
                type_=forms_xso.FieldType.JID_SINGLE,
                values=[],
                desc="some description",
            )
        )

        data.fields.append(
            forms_xso.Field(
                var="foo",
                type_=forms_xso.FieldType.TEXT_SINGLE,
                values=[],
            )
        )

        class F(form.Form):
            jid = fields.JIDSingle(
                var="jid",
                label="Foobar"
            )

        f = F.from_xso(data)
        f.jid.value = aioxmpp.JID.fromstr("foo@bar.baz")

        result = f.render_reply()
        self.assertIsInstance(result, forms_xso.Data)
        self.assertEqual(
            len(result.fields),
            2
        )

        self.assertIs(
            data.fields[1],
            result.fields[1]
        )

        self.assertIsNot(
            data.fields[0],
            result.fields[0],
        )

        jid_field = result.fields[0]
        self.assertSequenceEqual(
            jid_field.values,
            ["foo@bar.baz"]
        )
        self.assertEqual(
            jid_field.desc,
            "some description",
        )
        self.assertIsNone(
            jid_field.label
        )

    def test_render_reply_emits_FORM_TYPE(self):
        data = forms_xso.Data(type_=forms_xso.DataType.FORM)

        data.fields.append(
            forms_xso.Field(
                var="FORM_TYPE",
                type_=forms_xso.FieldType.HIDDEN,
                values=["some-uri"],
            )
        )

        data.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.FIXED,
                values=["This is some heading."],
            )
        )

        data.fields.append(
            forms_xso.Field(
                var="jid",
                type_=forms_xso.FieldType.JID_SINGLE,
                values=[],
                desc="some description",
            )
        )

        class F(form.Form):
            FORM_TYPE = "some-uri"

            jid = fields.JIDSingle(
                var="jid",
                label="Foobar"
            )

            other = fields.TextSingle(
                var="foo",
            )

        f = F.from_xso(data)
        f.jid.value = aioxmpp.JID.fromstr("foo@bar.baz")

        result = f.render_reply()
        self.assertIsInstance(result, forms_xso.Data)
        self.assertEqual(
            result.type_,
            forms_xso.DataType.SUBMIT,
        )
        self.assertEqual(
            len(result.fields),
            3
        )

        self.assertIs(
            data.fields[0],
            result.fields[0]
        )
        self.assertIs(
            data.fields[1],
            result.fields[1],
        )

        self.assertIsNot(
            data.fields[2],
            result.fields[2],
        )

        jid_field = result.fields[2]
        self.assertSequenceEqual(
            jid_field.values,
            ["foo@bar.baz"]
        )
        self.assertEqual(
            jid_field.desc,
            "some description",
        )
        self.assertIsNone(
            jid_field.label
        )

        self.assertEqual(
            data.get_form_type(),
            F.FORM_TYPE,
        )

    def test_render_request(self):
        class F(form.Form):
            jid = fields.JIDSingle(
                var="jid",
                required=True,
                desc="Enter a valid JID here",
                label="Your JID",
            )

            something_else = fields.TextSingle(
                var="other",
                label="Something else",
            )

        f = F()
        f.jid.value = aioxmpp.JID.fromstr("foo@bar.baz")
        f.something_else.value = "some_text"

        data = f.render_request()

        self.assertIsInstance(
            data,
            forms_xso.Data,
        )

        self.assertEqual(
            len(data.fields),
            2
        )

        for field in data.fields:
            self.assertIsInstance(
                field,
                forms_xso.Field,
            )

        jid_field = [field for field in data.fields
                     if field.var == "jid"].pop()
        self.assertEqual(
            jid_field.type_,
            forms_xso.FieldType.JID_SINGLE,
        )
        self.assertEqual(
            jid_field.label,
            "Your JID"
        )
        self.assertEqual(
            jid_field.desc,
            "Enter a valid JID here",
        )
        self.assertIs(
            jid_field.required,
            True
        )
        self.assertSequenceEqual(
            jid_field.values,
            ["foo@bar.baz"],
        )

        other_field = [field for field in data.fields
                       if field.var == "other"].pop()
        self.assertEqual(
            other_field.type_,
            forms_xso.FieldType.TEXT_SINGLE,
        )
        self.assertEqual(
            other_field.label,
            "Something else"
        )
        self.assertIsNone(
            other_field.desc,
        )
        self.assertIs(
            other_field.required,
            False,
        )
        self.assertSequenceEqual(
            other_field.values,
            ["some_text"],
        )

    def test_render_request_includes_FORM_TYPE(self):
        class F(form.Form):
            FORM_TYPE = "some-uri"

            jid = fields.JIDSingle(
                var="jid",
                required=True,
                desc="Enter a valid JID here",
                label="Your JID",
            )

            something_else = fields.TextSingle(
                var="other",
                label="Something else",
            )

        f = F()
        f.jid.value = aioxmpp.JID.fromstr("foo@bar.baz")
        f.something_else.value = "some_text"

        data = f.render_request()

        self.assertEqual(
            data.get_form_type(),
            f.FORM_TYPE,
        )

    def test_render_request_with_layout(self):
        class F(form.Form):
            jid = fields.JIDSingle(
                var="jid",
                required=True,
                desc="Enter a valid JID here",
                label="Your JID",
            )

            something_else = fields.TextSingle(
                var="other",
                label="Something else",
            )

            LAYOUT = [
                "Metadata",
                jid,
                "Captcha",
                something_else,
            ]

        f = F()
        f.jid.value = aioxmpp.JID.fromstr("foo@bar.baz")
        f.something_else.value = "some_text"

        data = f.render_request()

        self.assertIsInstance(
            data,
            forms_xso.Data,
        )

        self.assertEqual(
            len(data.fields),
            4
        )

        for field in data.fields:
            self.assertIsInstance(
                field,
                forms_xso.Field,
            )

        text_field = data.fields[0]
        self.assertEqual(
            text_field.type_,
            forms_xso.FieldType.FIXED,
        )
        self.assertIsNone(text_field.var)
        self.assertSequenceEqual(
            text_field.values,
            ["Metadata"]
        )

        jid_field = data.fields[1]
        self.assertEqual(
            jid_field.type_,
            forms_xso.FieldType.JID_SINGLE,
        )
        self.assertEqual(
            jid_field.label,
            "Your JID"
        )
        self.assertEqual(
            jid_field.desc,
            "Enter a valid JID here",
        )
        self.assertIs(
            jid_field.required,
            True
        )
        self.assertSequenceEqual(
            jid_field.values,
            ["foo@bar.baz"],
        )

        text_field = data.fields[2]
        self.assertEqual(
            text_field.type_,
            forms_xso.FieldType.FIXED,
        )
        self.assertIsNone(text_field.var)
        self.assertSequenceEqual(
            text_field.values,
            ["Captcha"]
        )

        other_field = data.fields[3]
        self.assertEqual(
            other_field.type_,
            forms_xso.FieldType.TEXT_SINGLE,
        )
        self.assertEqual(
            other_field.label,
            "Something else"
        )
        self.assertIsNone(
            other_field.desc,
        )
        self.assertIs(
            other_field.required,
            False,
        )
        self.assertSequenceEqual(
            other_field.values,
            ["some_text"],
        )



########################################################################
# File name: __main__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import gc
import sys

try:
    import quamash
    import PyQt5.Qt as Qt
except ImportError as exc:
    print(exc, file=sys.stderr)
    print("This example requires quamash and PyQt5.", file=sys.stderr)

from adhoc_browser.main import AdHocBrowser

qapp = Qt.QApplication(sys.argv)
qapp.setQuitOnLastWindowClosed(False)
asyncio.set_event_loop(quamash.QEventLoop(app=qapp))
loop = asyncio.get_event_loop()
try:
    example = AdHocBrowser()
    example.prepare_argparse()
    example.configure()
    loop.run_until_complete(example.run_example())
finally:
    loop.close()
asyncio.set_event_loop(None)
del example, loop, qapp
gc.collect()



########################################################################
# File name: test_oob.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.misc as misc_xso
import aioxmpp.xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_namespace(self):
        self.assertEqual(
            namespaces.xep0066_oob_x,
            "jabber:x:oob",
        )


class TestOOBExtension(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            misc_xso.OOBExtension,
            aioxmpp.xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            misc_xso.OOBExtension.TAG,
            (namespaces.xep0066_oob_x, "x"),
        )

    def test_url(self):
        self.assertIsInstance(
            misc_xso.OOBExtension.url,
            aioxmpp.xso.ChildText,
        )
        self.assertEqual(
            misc_xso.OOBExtension.url.tag,
            (namespaces.xep0066_oob_x, "url")
        )

    def test_message_hook(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0066_oob,
            aioxmpp.xso.Child,
        )
        self.assertIn(
            misc_xso.OOBExtension,
            aioxmpp.Message.xep0066_oob._classes,
        )



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.chatstates` – Chat State Notification support (:xep:`0085`)
##########################################################################

This module provides support to implement :xep:`Chat State
Notifications <85>`.

XSOs
====

The module registers an attribute ``xep0085_chatstate`` with
:class:`aioxmpp.Message` to represent the chat state
notification tags, it takes values from the following enumeration (or
:data:`None` if no tag is present):

.. autoclass:: ChatState

Helpers
=======

The module provides the following helper class, that handles the state
management for chat state notifications:

.. autoclass:: ChatStateManager

Its operation is controlled by one of the chat state strategies:

.. autoclass:: DoNotEmit

.. autoclass:: DiscoverSupport

.. autoclass:: AlwaysEmit

"""
from .xso import ChatState  # NOQA
from .utils import (ChatStateManager, DoNotEmit, AlwaysEmit,  # NOQA
                    DiscoverSupport)



########################################################################
# File name: test_cache.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import io
import unittest
import random

import aioxmpp.cache

from aioxmpp.benchtest import times, timed, record

class TestLRUDict(unittest.TestCase):
    KEY = "aioxmpp.cache", "LRUDict"

    @times(1000)
    def test_random_access(self):
        key = self.KEY + ("random_access",)

        N = 1000

        lru_dict = aioxmpp.cache.LRUDict()
        lru_dict.maxsize = N
        keys = [object() for i in range(N)]
        for i in range(N):
            lru_dict[keys[i]] = object()

        with timed() as t:
            for i in range(N):
                lru_dict[keys[random.randrange(0, N)]]

        record(key, t.elapsed, "s")

    @times(1000)
    def test_inserts(self):
        key = self.KEY + ("inserts",)

        N = 1000

        lru_dict = aioxmpp.cache.LRUDict()
        lru_dict.maxsize = N
        keys = [object() for i in range(N)]
        for i in range(N):
            lru_dict[keys[i]] = object()

        with timed() as t:
            for i in range(N):
                lru_dict[object()] = object()

        record(key, t.elapsed, "s")



########################################################################
# File name: body.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import functools

import aioxmpp.structs


@functools.singledispatch
def set_message_body(x, message):
    raise NotImplementedError(
        "type {!r} is not supported as message body".format(
            type(x)
        )
    )


@set_message_body.register(str)
def set_message_body_str(x, message):
    message.body.clear()
    message.body[None] = x


@set_message_body.register(aioxmpp.structs.LanguageMap)
def set_message_body_langmap(x, message):
    message.body.clear()
    message.body.update(x)



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest
import unittest.mock

import aioxmpp.xso as xso

import aioxmpp.rsm.xso as rsm_xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_rsm(self):
        self.assertEqual(
            namespaces.xep0059_rsm,
            "http://jabber.org/protocol/rsm"
        )


class Test_RangeLimitBase(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            rsm_xso._RangeLimitBase,
            xso.XSO,
        ))

    def test_value(self):
        self.assertIsInstance(
            rsm_xso._RangeLimitBase.value,
            xso.Text,
        )
        self.assertIsNone(
            rsm_xso._RangeLimitBase.value.default,
        )

    def test_init_default(self):
        l = rsm_xso._RangeLimitBase()
        self.assertIsNone(l.value)

    def test_init(self):
        l = rsm_xso._RangeLimitBase("foo")
        self.assertEqual(l.value, "foo")


class TestAfter(unittest.TestCase):
    def test_is_range_limit(self):
        self.assertTrue(issubclass(
            rsm_xso.After,
            rsm_xso._RangeLimitBase,
        ))

    def test_tag(self):
        self.assertEqual(
            rsm_xso.After.TAG,
            (namespaces.xep0059_rsm, "after"),
        )


class TestFirst(unittest.TestCase):
    def test_is_range_limit(self):
        self.assertTrue(issubclass(
            rsm_xso.First,
            rsm_xso._RangeLimitBase,
        ))

    def test_tag(self):
        self.assertEqual(
            rsm_xso.First.TAG,
            (namespaces.xep0059_rsm, "first"),
        )

    def test_index(self):
        self.assertIsInstance(
            rsm_xso.First.index,
            xso.Attr,
        )
        self.assertEqual(
            rsm_xso.First.index.tag,
            (None, "index"),
        )
        self.assertIsInstance(
            rsm_xso.First.index.type_,
            xso.Integer,
        )
        self.assertIsNone(rsm_xso.First.index.default)


class TestLast(unittest.TestCase):
    def test_is_range_limit(self):
        self.assertTrue(issubclass(
            rsm_xso.Last,
            rsm_xso._RangeLimitBase,
        ))

    def test_tag(self):
        self.assertEqual(
            rsm_xso.Last.TAG,
            (namespaces.xep0059_rsm, "last"),
        )


class TestBefore(unittest.TestCase):
    def test_is_range_limit(self):
        self.assertTrue(issubclass(
            rsm_xso.Before,
            rsm_xso._RangeLimitBase,
        ))

    def test_tag(self):
        self.assertEqual(
            rsm_xso.Before.TAG,
            (namespaces.xep0059_rsm, "before"),
        )


class TestResultSetMetadata(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            rsm_xso.ResultSetMetadata,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            rsm_xso.ResultSetMetadata.TAG,
            (namespaces.xep0059_rsm, "set"),
        )

    def test_before(self):
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.before,
            xso.Child,
        )
        self.assertSetEqual(
            rsm_xso.ResultSetMetadata.before._classes,
            {
                rsm_xso.Before,
            }
        )

    def test_after(self):
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.after,
            xso.Child,
        )
        self.assertSetEqual(
            rsm_xso.ResultSetMetadata.after._classes,
            {
                rsm_xso.After,
            }
        )

    def test_first(self):
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.first,
            xso.Child,
        )
        self.assertSetEqual(
            rsm_xso.ResultSetMetadata.first._classes,
            {
                rsm_xso.First,
            }
        )

    def test_last(self):
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.last,
            xso.Child,
        )
        self.assertSetEqual(
            rsm_xso.ResultSetMetadata.last._classes,
            {
                rsm_xso.Last,
            }
        )

    def test_count(self):
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.count,
            xso.ChildText,
        )
        self.assertEqual(
            rsm_xso.ResultSetMetadata.count.tag,
            (namespaces.xep0059_rsm, "count"),
        )
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.count.type_,
            xso.Integer,
        )
        self.assertIsNone(rsm_xso.ResultSetMetadata.count.default)

    def test_index(self):
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.index,
            xso.ChildText,
        )
        self.assertEqual(
            rsm_xso.ResultSetMetadata.index.tag,
            (namespaces.xep0059_rsm, "index"),
        )
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.index.type_,
            xso.Integer,
        )
        self.assertIsNone(rsm_xso.ResultSetMetadata.index.default)

    def test_max(self):
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.max_,
            xso.ChildText,
        )
        self.assertEqual(
            rsm_xso.ResultSetMetadata.max_.tag,
            (namespaces.xep0059_rsm, "max"),
        )
        self.assertIsInstance(
            rsm_xso.ResultSetMetadata.max_.type_,
            xso.Integer,
        )
        self.assertIsNone(rsm_xso.ResultSetMetadata.max_.default)

    def test_init(self):
        rsm = rsm_xso.ResultSetMetadata()
        self.assertIsNone(rsm.first)
        self.assertIsNone(rsm.last)
        self.assertIsNone(rsm.after)
        self.assertIsNone(rsm.before)
        self.assertIsNone(rsm.count)
        self.assertIsNone(rsm.max_)
        self.assertIsNone(rsm.index)

    def test_fetch_page(self):
        rsm = rsm_xso.ResultSetMetadata.fetch_page(10)
        self.assertEqual(rsm.index, 10)
        self.assertIsNone(rsm.max_)

    def test_fetch_page_with_count(self):
        rsm = rsm_xso.ResultSetMetadata.fetch_page(20, max_=100)
        self.assertEqual(rsm.index, 20)
        self.assertEqual(rsm.max_, 100)

    def test_limit_cls(self):
        rsm = rsm_xso.ResultSetMetadata.limit(100)
        self.assertEqual(rsm.max_, 100)

    def test_limit_index_obj(self):
        rsm = rsm_xso.ResultSetMetadata.fetch_page(10)
        new_rsm = rsm.limit(100)
        self.assertIsNot(rsm, new_rsm)
        self.assertEqual(rsm.index, new_rsm.index)
        self.assertEqual(new_rsm.max_, 100)
        self.assertIsNone(rsm.max_)

    def test_limit_before_after_obj(self):
        rsm = rsm_xso.ResultSetMetadata()
        rsm.after = rsm_xso.After()
        rsm.before = rsm_xso.Before()
        new_rsm = rsm.limit(100)
        self.assertIsNot(rsm, new_rsm)
        self.assertIsNot(rsm.before, new_rsm.before)
        self.assertIsNot(rsm.after, new_rsm.after)
        self.assertEqual(rsm.after.value, new_rsm.after.value)
        self.assertEqual(rsm.before.value, new_rsm.before.value)
        self.assertEqual(new_rsm.max_, 100)
        self.assertIsNone(rsm.max_)

    def test_next__obj(self):
        rsm = rsm_xso.ResultSetMetadata()
        rsm.last = unittest.mock.Mock()
        rsm.last.value = "last value"

        new_rsm = rsm.next_page()
        self.assertIsNot(rsm, new_rsm)
        self.assertIsNot(rsm.last, new_rsm.after)
        self.assertEqual(rsm.last.value, new_rsm.after.value)
        self.assertIsNone(new_rsm.max_)

    def test_next__with_max_obj(self):
        rsm = rsm_xso.ResultSetMetadata()
        rsm.last = unittest.mock.Mock()
        rsm.last.value = "last value"

        new_rsm = rsm.next_page(max_=100)
        self.assertIsNot(rsm, new_rsm)
        self.assertIsNot(rsm.last, new_rsm.after)
        self.assertEqual(rsm.last.value, new_rsm.after.value)
        self.assertEqual(new_rsm.max_, 100)

    def test_previous_obj(self):
        rsm = rsm_xso.ResultSetMetadata()
        rsm.first = unittest.mock.Mock()
        rsm.first.value = "first value"

        new_rsm = rsm.previous_page()
        self.assertIsNot(rsm, new_rsm)
        self.assertIsNot(rsm.first, new_rsm.before)
        self.assertEqual(rsm.first.value, new_rsm.before.value)
        self.assertIsNone(new_rsm.max_)

    def test_previous_with_max_obj(self):
        rsm = rsm_xso.ResultSetMetadata()
        rsm.first = unittest.mock.Mock()
        rsm.first.value = "first value"

        new_rsm = rsm.previous_page(max_=100)
        self.assertIsNot(rsm, new_rsm)
        self.assertIsNot(rsm.first, new_rsm.before)
        self.assertEqual(rsm.first.value, new_rsm.before.value)
        self.assertEqual(new_rsm.max_, 100)

    def test_last_cls(self):
        rsm = rsm_xso.ResultSetMetadata.last_page()
        self.assertIsInstance(rsm.before, rsm_xso.Before)
        self.assertIsNone(rsm.before.value)
        self.assertIsNone(rsm.max_)

    def test_last_with_max_cls(self):
        rsm = rsm_xso.ResultSetMetadata.last_page(max_=10)
        self.assertIsInstance(rsm.before, rsm_xso.Before)
        self.assertIsNone(rsm.before.value)
        self.assertEqual(rsm.max_, 10)



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.xso` --- Working with XML stream contents
########################################################

This subpackage deals with **X**\ ML **S**\ tream **O**\ bjects. XSOs can be
stanzas, but in general any XML.

The facilities in this subpackage are supposed to help developers of XEP
plugins, as well as the main development of :mod:`aioxmpp`. The subpackage
is split in two parts, :mod:`aioxmpp.xso.model`, which provides facilities to
allow declarative-style parsing and un-parsing of XML subtrees into XSOs and
the :mod:`aioxmpp.xso.types` module, which provides classes which implement
validators and type parsers for content represented as strings in XML.

Terminology
===========

Definition of an XSO
--------------------

An XSO is an object whose class inherits from
:class:`aioxmpp.xso.XSO`.

A word on tags
--------------

Tags, as used by etree, are used throughout this module. Note that we are
representing tags as tuples of ``(namespace_uri, localname)``, where
``namespace_uri`` may be :data:`None`.

.. seealso::

   The functions :func:`normalize_tag` and :func:`tag_to_str` are useful to
   convert from and to ElementTree compatible strings.


Suspendable functions
---------------------

This module uses suspendable functions, implemented as generators, at several
points. These may also be called coroutines, but have nothing to do with
coroutines as used by :mod:`asyncio`, which is why we will call them
suspendable functions here.

Suspendable functions possibly take arguments and then operate on input which
is fed to them in a push-manner step by step (using the
:meth:`~types.GeneratorType.send` method). The main usage in this module is to
process SAX events: The SAX events are processed step-by-step by the functions,
and when the event is fully processed, it suspends itself (using ``yield``)
until the next event is sent into it.

General functions
=================

.. autofunction:: normalize_tag

.. autofunction:: tag_to_str

.. module:: aioxmpp.xso.model

.. currentmodule:: aioxmpp.xso

Object declaration with :mod:`aioxmpp.xso.model`
================================================

This module provides facilities to create classes which map to full XML stream
subtrees (for example stanzas including payload).

To create such a class, derive from :class:`XSO` and provide attributes
using the :class:`Attr`, :class:`Text`, :class:`Child` and :class:`ChildList`
descriptors.

Descriptors for XML-sourced attributes
--------------------------------------

The following descriptors can be used to load XSO attributes from XML. There
are two fundamentally different descriptor types: *scalar* and *non-scalar*
(e.g. list) descriptors. Assignment to the descriptor attribute is
strictly type-checked for *scalar* descriptors.

Scalar descriptors
^^^^^^^^^^^^^^^^^^

Many of the arguments and attributes used for the scalar descriptors are
similar. They are described in detail on the :class:`Attr` class and not
repeated that detailed on the other classes. Refer to the documentation of the
:class:`Attr` class in those cases.

.. autoclass:: Attr(name, *[, type_=xso.String()][, validator=None][, validate=ValidateMode.FROM_RECV][, missing=None][, default][, erroneous_as_absent=False])

.. autoclass:: LangAttr(*[, validator=None][, validate=ValidateMode.FROM_RECV][, default=None])

.. autoclass:: Child(classes, *[, required=False][, strict=False])

.. autoclass:: ChildTag(tags, *[, text_policy=UnknownTextPolicy.FAIL][, child_policy=UnknownChildPolicy.FAIL][, attr_policy=UnknownAttrPolicy.FAIL][, default_ns=None][, allow_none=False])

.. autoclass:: ChildFlag(tag, *[, text_policy=UnknownTextPolicy.FAIL][, child_policy=UnknownChildPolicy.FAIL][, attr_policy=UnknownAttrPolicy.FAIL])

.. autoclass:: ChildText(tag, *[, child_policy=UnknownChildPolicy.FAIL][, attr_policy=UnknownAttrPolicy.FAIL][, type_=xso.String()][, validator=None][, validate=ValidateMode.FROM_RECV][, default][, erroneous_as_absent=False])

.. autoclass:: Text(*[, type_=xso.String()][, validator=None][, validate=ValidateMode.FROM_RECV][, default][, erroneous_as_absent=False])

Non-scalar descriptors
^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: ChildList(classes)

.. autoclass:: ChildMap(classes[, key=None])

.. autoclass:: ChildLangMap(classes)

.. autoclass:: ChildValueList(type_)

.. autoclass:: ChildValueMap(type_, *, mapping_type=dict)

.. autoclass:: ChildValueMultiMap(type_, *, mapping_type=multidict.MultiDict)

.. autoclass:: ChildTextMap(xso_type)

.. autoclass:: Collector()

Container for child lists
^^^^^^^^^^^^^^^^^^^^^^^^^

The child lists in :class:`ChildList`, :class:`ChildMap` and
:class:`ChildLangMap` descriptors use a specialized list-subclass which
provides advanced capabilities for filtering :class:`XSO` objects.

.. currentmodule:: aioxmpp.xso.model

.. autoclass:: XSOList

.. currentmodule:: aioxmpp.xso


Parsing XSOs
------------

To parse XSOs, an asynchronous approach which uses SAX-like events is
followed. For this, the suspendable functions explained earlier are used. The
main class to parse a XSO from events is :class:`XSOParser`. To drive
that suspendable callable from SAX events, use a :class:`SAXDriver`.

.. autoclass:: XSOParser

.. autoclass:: SAXDriver

Base and meta class
-------------------

The :class:`XSO` base class makes use of the :class:`model.XMLStreamClass`
metaclass and provides implementations for utility methods. For an object to
work with this module, it must derive from :class:`XSO` or provide an
identical interface.

.. autoclass:: XSO()

.. autoclass:: CapturingXSO()

The metaclass takes care of collecting the special descriptors in attributes
where they can be used by the SAX event interpreter to fill the class with
data. It also provides a class method for late registration of child classes.

.. currentmodule:: aioxmpp.xso.model

.. autoclass:: XMLStreamClass

.. currentmodule:: aioxmpp.xso

To create an enumeration of XSO classes, the following mixin can be used:

.. autoclass:: XSOEnumMixin

Functions, enumerations and exceptions
--------------------------------------

The values of the following enumerations are used on "magic" attributes of
:class:`XMLStreamClass` instances (i.e. classes).

.. autoclass:: UnknownChildPolicy

.. autoclass:: UnknownAttrPolicy

.. autoclass:: UnknownTextPolicy

.. autoclass:: ValidateMode

The following exceptions are generated at some places in this module:

.. autoclass:: UnknownTopLevelTag

The following special value is used to indicate that no default is used with a
descriptor:

.. data:: NO_DEFAULT

   This is a special value which is used to indicate that no defaulting should
   take place. It can be passed to the `default` arguments of descriptors, and
   usually is the default value of these arguments.

   It compares unequal to everything but itself, does not support ordering,
   conversion to bool, float or integer.

.. autofunction:: capture_events

.. autofunction:: events_to_sax

Handlers for missing attributes
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autofunction:: lang_attr

.. module:: aioxmpp.xso.types

.. currentmodule:: aioxmpp.xso

Types and validators from :mod:`~aioxmpp.xso.types`
===================================================

This module provides classes whose objects can be used as types and validators
in :mod:`~aioxmpp.xso.model`.

Character Data types
--------------------

These types describe character data, i.e. text in XML. Thus, they can be used
with :class:`Attr`, :class:`Text` and similar descriptors. They are used to
deserialise XML character data to python values, such as integers or dates and
vice versa. These types inherit from :class:`AbstractCDataType`.

.. autoclass:: String

.. autoclass:: Integer

.. autoclass:: Bool

.. autoclass:: DateTime

.. autoclass:: Date

.. autoclass:: Time

.. autoclass:: Base64Binary

.. autoclass:: HexBinary

.. autoclass:: JID

.. autoclass:: ConnectionLocation

.. autoclass:: LanguageTag

.. autoclass:: EnumCDataType(enum_class, nested_type=xso.String(), *, allow_coerce=False, deprecate_coerce=False, allow_unknown=True, accept_unknown=True)

.. autofunction:: EnumType(enum_class[, nested_type], *, allow_coerce=False, deprecate_coerce=False, allow_unknown=True, accept_unknown=True)

.. autoclass:: Unknown

Element types
-------------

These types describe structured XML data, i.e. subtrees. Thus, they can be used
with the :class:`ChildValueList` and :class:`ChildValueMap` family of
descriptors (which represent XSOs as python values). These types inherit from
:class:`AbstractElementType`.

.. autoclass:: EnumElementType

.. autoclass:: TextChildMap

Defining custom types
---------------------

.. autoclass:: AbstractCDataType

.. autoclass:: AbstractElementType

Validators
----------

Validators validate the python values after they have been parsed from
XML-sourced strings or even when being assigned to a descriptor attribute
(depending on the choice in the `validate` argument).

They can be useful both for defending and rejecting incorrect input and to
avoid producing incorrect output.

The basic validator interface
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: AbstractValidator

Implementations
^^^^^^^^^^^^^^^

.. autoclass:: RestrictToSet

.. autoclass:: Nmtoken

.. autoclass:: IsInstance

.. autoclass:: NumericRange

.. module:: aioxmpp.xso.query

.. currentmodule:: aioxmpp.xso

Querying data from XSOs
=======================

With XML, we have XPath as query language to retrieve data from XML trees. With
XSOs, we have :mod:`aioxmpp.xso.query`, even though it’s not as powerful as
XPath.

Syntactically, it’s oriented on XPath. Consider the following XSO classes:

.. code-block:: python

    class FooXSO(xso.XSO):
        TAG = (None, "foo")

        attr = xso.Attr(
            "attr"
        )


    class BarXSO(xso.XSO):
        TAG = (None, "bar")

        child = xso.Child([
            FooXSO,
        ])


    class BazXSO(FooXSO):
        TAG = (None, "baz")

        attr2 = xso.Attr(
            "attr2"
        )


    class RootXSO(xso.XSO):
        TAG = (None, "root")

        children = xso.ChildList([
            FooXSO,
            BarXSO,
        ])

        attr = xso.Attr(
            "attr"
        )


To perform a query, we first need to set up a
:class:`.query.EvaluationContext`:

.. code-block:: python

   root_xso = # a RootXSO instance
   ec = xso.query.EvaluationContext()
   ec.set_toplevel_object(root_xso)

Using the context, we can now execute queries:

.. code-block:: python

   # to find all FooXSO children of the RootXSO
   ec.eval(RootXSO.children / FooXSO)

   # to find all BarXSO children of the RootXSO
   ec.eval(RootXSO.children / BarXSO)

   # to find all FooXSO children of the RootXSO, where FooXSO.attr
   # is set
   ec.eval(RootXSO.children / FooXSO[where(FooXSO.attr)])

   # to find all FooXSO children of the RootXSO, where FooXSO.attr
   # is *not* set
   ec.eval(RootXSO.children / FooXSO[where(not FooXSO.attr)])

   # to find all FooXSO children of the RootXSO, where FooXSO.attr
   # is set to "foobar"
   ec.eval(RootXSO.children / FooXSO[where(FooXSO.attr == "foobar")])

   # to test whether there is a FooXSO which has attr set to
   # "foobar"
   ec.eval(RootXSO.children / FooXSO.attr == "foobar")

   # to find the first three FooXSO children where attr is set
   ec.eval(RootXSO.children / FooXSO[where(FooXSO.attr)][:3])

The following operators are available in the :mod:`aioxmpp.xso` namespace:

.. autoclass:: where

.. autofunction:: not_

The following need to be explicitly sourced from :mod:`aioxmpp.xso.query`, as
they are rarely used directly in user code.

.. currentmodule:: aioxmpp.xso.query

.. autoclass:: EvaluationContext()

.. note::

   The implementation details of the query language are documented in the
   source. They are not useful unless you want to implement custom query
   operators, which is not possible without modifying the
   :mod:`aioxmpp.xso.query` source anyways.

.. currentmodule:: aioxmpp.xso

Predefined XSO base classes
===========================

Some patterns reoccur when using this subpackage. For these, base classes are
provided which faciliate the use.

.. autoclass:: AbstractTextChild


"""


def tag_to_str(tag):
    """
    `tag` must be a tuple ``(namespace_uri, localname)``. Return a tag string
    conforming to the ElementTree specification. Example::

         tag_to_str(("jabber:client", "iq")) == "{jabber:client}iq"
    """
    return "{{{:s}}}{:s}".format(*tag) if tag[0] else tag[1]


def normalize_tag(tag):
    """
    Normalize an XML element tree `tag` into the tuple format. The following
    input formats are accepted:

    * ElementTree namespaced string, e.g. ``{uri:bar}foo``
    * Unnamespaced tags, e.g. ``foo``
    * Two-tuples consisting of `namespace_uri` and `localpart`; `namespace_uri`
      may be :data:`None` if the tag is supposed to be namespaceless. Otherwise
      it must be, like `localpart`, a :class:`str`.

    Return a two-tuple consisting the ``(namespace_uri, localpart)`` format.
    """
    if isinstance(tag, str):
        namespace_uri, sep, localname = tag.partition("}")
        if sep:
            if not namespace_uri.startswith("{"):
                raise ValueError("not a valid etree-format tag")
            namespace_uri = namespace_uri[1:]
        else:
            localname = namespace_uri
            namespace_uri = None
        return (namespace_uri, localname)
    elif len(tag) != 2:
        raise ValueError("not a valid tuple-format tag")
    else:
        if any(part is not None and not isinstance(part, str) for part in tag):
            raise TypeError("tuple-format tags must only contain str and None")
        if tag[1] is None:
            raise ValueError("tuple-format localname must not be None")
    return tag


from .types import (  # NOQA
    Unknown,
    AbstractCDataType,
    AbstractElementType,
    String,
    Integer,
    Float,
    Bool,
    DateTime,
    Date,
    Time,
    Base64Binary,
    HexBinary,
    JID,
    ConnectionLocation,
    LanguageTag,
    TextChildMap,
    EnumType,
    EnumCDataType,
    EnumElementType,
    AbstractValidator,
    RestrictToSet,
    Nmtoken,
    IsInstance,
    NumericRange,
)

from .model import (  # NOQA
    tag_to_str,
    normalize_tag,
    UnknownChildPolicy,
    UnknownAttrPolicy,
    UnknownTextPolicy,
    ValidateMode,
    UnknownTopLevelTag,
    Attr,
    LangAttr,
    Child,
    ChildFlag,
    ChildList,
    ChildLangMap,
    ChildMap,
    ChildTag,
    ChildText,
    Collector,
    Text,
    ChildValueList,
    ChildValueMap,
    ChildValueMultiMap,
    ChildTextMap,
    XSOParser,
    SAXDriver,
    XSO,
    XSOEnumMixin,
    CapturingXSO,
    lang_attr,
    capture_events,
    events_to_sax,
)


class AbstractTextChild(XSO):
    """
    One of the recurring patterns when using :mod:`xso` is the use of a XSO
    subclass to represent an XML node which has only character data and an
    ``xml:lang`` attribute.

    The `text` and `lang` arguments to the constructor can be used to
    initialize the attributes.

    This class provides exactly that. It inherits from :class:`XSO`.

    .. attribute:: lang

       The ``xml:lang`` of the node, as :class:`~.structs.LanguageTag`.

    .. attribute:: text

       The textual content of the node (XML character data).

    Example use as base class::

      class Subject(xso.AbstractTextChild):
          TAG = (namespaces.client, "subject")

    The full example can also be found in the source code of
    :class:`.stanza.Subject`.

    """

    lang = LangAttr()
    text = Text(default=None)

    def __init__(self, text=None, lang=None):
        super().__init__()
        self.text = text
        self.lang = lang

    def __eq__(self, other):
        try:
            other_key = (other.lang, other.text)
        except AttributeError:
            return NotImplemented
        return (self.lang, self.text) == other_key

from .model import _PropBase
NO_DEFAULT = _PropBase.NO_DEFAULT
del _PropBase


from .query import (  # NOQA
    where,
    not_,
)



########################################################################
# File name: test_e2e.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp

from aioxmpp.e2etest import (
    TestCase,
    blocking_timed,
    blocking,
)


class TestPresence(TestCase):
    @blocking
    @asyncio.coroutine
    def setUp(self):
        self.a, self.b = yield from asyncio.gather(
            self.provisioner.get_connected_client(),
            self.provisioner.get_connected_client(),
        )

    @blocking_timed
    @asyncio.coroutine
    def test_events_on_presence(self):
        pres = aioxmpp.Presence(
            to=self.b.local_jid.bare(),
            type_=aioxmpp.PresenceType.AVAILABLE,
        )

        svc = self.b.summon(aioxmpp.PresenceClient)

        bare_fut = asyncio.Future()
        avail_fut = asyncio.Future()
        changed_fut = asyncio.Future()

        def on_bare_available(stanza):
            if stanza.from_ == self.b.local_jid:
                return False
            bare_fut.set_result(stanza)
            return True

        def on_available(full_jid, stanza):
            if full_jid == self.b.local_jid:
                return False
            avail_fut.set_result((full_jid, stanza))
            return True

        def on_changed(full_jid, stanza):
            changed_fut.set_result((full_jid, stanza))
            return True

        svc.on_bare_available.connect(on_bare_available)
        svc.on_available.connect(on_available)
        svc.on_changed.connect(on_changed)

        yield from self.a.send(pres)

        stanza = yield from bare_fut
        self.assertIsInstance(stanza, aioxmpp.Presence)
        self.assertEqual(stanza.type_, aioxmpp.PresenceType.AVAILABLE)
        self.assertEqual(stanza.from_, self.a.local_jid)

        full_jid, stanza = yield from avail_fut
        self.assertIsInstance(stanza, aioxmpp.Presence)
        self.assertEqual(stanza.type_, aioxmpp.PresenceType.AVAILABLE)
        self.assertEqual(stanza.from_, self.a.local_jid)
        self.assertEqual(full_jid, stanza.from_)

        self.assertFalse(changed_fut.done())

        pres.show = aioxmpp.PresenceShow.DND
        yield from self.a.send(pres)

        full_jid, stanza = yield from changed_fut
        self.assertIsInstance(stanza, aioxmpp.Presence)
        self.assertEqual(stanza.type_, aioxmpp.PresenceType.AVAILABLE)
        self.assertEqual(stanza.show, aioxmpp.PresenceShow.DND)
        self.assertEqual(stanza.from_, self.a.local_jid)
        self.assertEqual(full_jid, stanza.from_)

    @blocking_timed
    @asyncio.coroutine
    def test_presence_bookkeeping(self):
        pres = aioxmpp.Presence(
            to=self.b.local_jid.bare(),
            type_=aioxmpp.PresenceType.AVAILABLE,
        )

        svc = self.b.summon(aioxmpp.PresenceClient)

        bare_fut = asyncio.Future()

        # we use the future as a synchronisation primitive to avoid reading
        # from the service before the stanza has actually been received
        def on_bare_available(stanza):
            if stanza.from_ == self.b.local_jid:
                return False
            bare_fut.set_result(None)
            return True

        svc.on_bare_available.connect(on_bare_available)

        yield from self.a.send(pres)

        yield from bare_fut

        most_available = svc.get_most_available_stanza(
            self.a.local_jid.bare()
        )
        self.assertIsNotNone(most_available)

        peer_resources = svc.get_peer_resources(
            self.a.local_jid.bare()
        )
        self.assertIn(
            self.a.local_jid.resource,
            peer_resources,
        )

        last_presence = svc.get_stanza(
            self.a.local_jid
        )
        self.assertIsNotNone(last_presence)



########################################################################
# File name: test_custom_queue.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import unittest

import aioxmpp.custom_queue as custom_queue

from aioxmpp.testutils import run_coroutine


class TestAsyncDeque(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.get_event_loop()
        self.q = custom_queue.AsyncDeque(loop=self.loop)

    def test_put_get_cycle_nowait(self):
        self.q.put_nowait(1)
        self.q.put_nowait(2)
        self.q.put_nowait(3)

        self.assertEqual(
            1,
            self.q.get_nowait()
        )
        self.assertEqual(
            2,
            self.q.get_nowait()
        )
        self.assertEqual(
            3,
            self.q.get_nowait()
        )

    def test_putleft_getright_cycle_nowait(self):
        self.q.putleft_nowait(1)
        self.q.putleft_nowait(2)
        self.q.putleft_nowait(3)

        self.assertEqual(
            1,
            self.q.getright_nowait()
        )
        self.assertEqual(
            2,
            self.q.getright_nowait()
        )
        self.assertEqual(
            3,
            self.q.getright_nowait()
        )

    def test_len(self):
        self.assertEqual(0, len(self.q))
        self.q.put_nowait(1)
        self.assertEqual(1, len(self.q))
        self.q.put_nowait(1)
        self.assertEqual(2, len(self.q))
        self.q.get_nowait()
        self.assertEqual(1, len(self.q))

    def test_contains(self):
        self.assertNotIn(1, self.q)
        self.q.put_nowait(1)
        self.assertIn(1, self.q)

    def test_get(self):
        @asyncio.coroutine
        def putter():
            yield from asyncio.sleep(0.001)
            self.q.put_nowait(1)

        _, v = run_coroutine(asyncio.gather(
            putter(),
            self.q.get()
        ))

        self.assertEqual(1, v)

    def test_one_producer_many_consumers(self):
        @asyncio.coroutine
        def putter():
            for i in range(20):
                self.q.put_nowait(i)
                yield from asyncio.sleep(0)

        @asyncio.coroutine
        def getter():
            result = []
            for i in range(4):
                result.append((yield from self.q.get()))
                yield from asyncio.sleep(0)
            return result

        _, vs1, vs2, vs3, vs4, vs5 = run_coroutine(asyncio.gather(
            putter(),
            getter(),
            getter(),
            getter(),
            getter(),
            getter(),
        ))

        self.assertSequenceEqual(
            range(20),
            sorted(vs1 + vs2 + vs3 + vs4 + vs5)
        )

    def test_one_consumer_many_producers(self):
        @asyncio.coroutine
        def putter(n0):
            for i in range(n0, n0 + 4):
                self.q.put_nowait(i)
                yield from asyncio.sleep(0)

        @asyncio.coroutine
        def getter():
            result = []
            for i in range(20):
                result.append((yield from self.q.get()))
                yield from asyncio.sleep(0)
            return result

        _, _, _, _, _, vs = run_coroutine(asyncio.gather(
            putter(0),
            putter(4),
            putter(8),
            putter(12),
            putter(16),
            getter()
        ))

        self.assertSequenceEqual(
            range(20),
            sorted(vs)
        )

    def test_empty(self):
        self.assertTrue(self.q.empty())
        self.q.put_nowait(1)
        self.assertFalse(self.q.empty())
        self.q.get_nowait()
        self.assertTrue(self.q.empty())

    def test_raise_asyncio_QueueEmpty_on_empty_get(self):
        with self.assertRaises(asyncio.QueueEmpty):
            self.q.get_nowait()

        with self.assertRaises(asyncio.QueueEmpty):
            self.q.getright_nowait()

    def test_clear(self):
        self.q.put_nowait(1)
        self.q.put_nowait(2)
        self.q.clear()
        self.assertTrue(self.q.empty())
        with self.assertRaises(asyncio.QueueEmpty):
            self.q.get_nowait()

    def tearDown(self):
        del self.q
        del self.loop



########################################################################
# File name: test_dispatcher.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import contextlib
import unittest
import unittest.mock

import aioxmpp
import aioxmpp.service
import aioxmpp.stream

import aioxmpp.dispatcher as dispatcher

from aioxmpp.testutils import (
    make_connected_client,
)


TEST_JID = aioxmpp.JID.fromstr("foo@bar.example/baz")
TEST_LOCAL_JID = aioxmpp.JID.fromstr("foo@local.example")


class FooStanza:
    def __init__(self, from_, type_):
        self.from_ = from_
        self.type_ = type_


class FooDispatcher(dispatcher.SimpleStanzaDispatcher):
    @property
    def local_jid(self):
        return TEST_LOCAL_JID


class TestSimpleStanzaDispatcher(unittest.TestCase):
    def setUp(self):
        self.d = FooDispatcher()

        self.handlers = unittest.mock.Mock()

        self.d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID,
            self.handlers.type_fulljid_no_wildcard,
            wildcard_resource=False,
        )

        self.d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID.bare(),
            self.handlers.type_barejid_no_wildcard,
            wildcard_resource=False,
        )

        self.d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID.bare(),
            self.handlers.type_barejid_wildcard,
            wildcard_resource=True,
        )

        self.d.register_callback(
            unittest.mock.sentinel.type_,
            None,
            self.handlers.type_wildcard,
            wildcard_resource=False,
        )

        self.d.register_callback(
            None,
            TEST_JID,
            self.handlers.wildcard_fulljid_no_wildcard,
            wildcard_resource=False,
        )

        self.d.register_callback(
            None,
            TEST_JID.bare(),
            self.handlers.wildcard_barejid_no_wildcard,
            wildcard_resource=False,
        )

        self.d.register_callback(
            None,
            TEST_JID.bare(),
            self.handlers.wildcard_barejid_wildcard,
            wildcard_resource=True,
        )

        self.d.register_callback(
            None,
            None,
            self.handlers.wildcard_wildcard,
            wildcard_resource=False,
        )

    def tearDown(self):
        del self.d

    def test_register_callback_rejects_dups(self):
        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID,
            unittest.mock.sentinel.cb,
        )

        with self.assertRaisesRegex(
                ValueError,
                "only one listener allowed"):
            d.register_callback(
                unittest.mock.sentinel.type_,
                TEST_JID,
                unittest.mock.sentinel.cb2,
            )

    def test_register_callback_flattens_wildcard_resource_for_fulljid(self):
        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID,
            unittest.mock.sentinel.cb,
            wildcard_resource=False,
        )

        with self.assertRaisesRegex(
                ValueError,
                "only one listener allowed"):
            d.register_callback(
                unittest.mock.sentinel.type_,
                TEST_JID,
                unittest.mock.sentinel.cb,
            )

    def test_register_callback_flattens_wildcard_resource_for_None(self):
        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            None,
            unittest.mock.sentinel.cb,
            wildcard_resource=False,
        )

        with self.assertRaisesRegex(
                ValueError,
                "only one listener allowed"):
            d.register_callback(
                unittest.mock.sentinel.type_,
                None,
                unittest.mock.sentinel.cb,
            )

    def test_register_callback_honors_wildcard_resource_for_bare(self):
        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID.bare(),
            unittest.mock.sentinel.cb,
            wildcard_resource=False,
        )

        d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID.bare(),
            unittest.mock.sentinel.cb,
            wildcard_resource=True,
        )

    def test_unregister_removes_callback(self):
        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID.bare(),
            unittest.mock.sentinel.cb,
            wildcard_resource=False,
        )

        d.unregister_callback(
            unittest.mock.sentinel.type_,
            TEST_JID.bare(),
            wildcard_resource=False,
        )

        d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID.bare(),
            unittest.mock.sentinel.cb,
            wildcard_resource=False,
        )

    def test_unregister_flattens_wildcard_resource_for_fulljid(self):
        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID,
            unittest.mock.sentinel.cb,
        )

        d.unregister_callback(
            unittest.mock.sentinel.type_,
            TEST_JID,
            wildcard_resource=False,
        )

        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID,
            unittest.mock.sentinel.cb,
        )

        d.unregister_callback(
            unittest.mock.sentinel.type_,
            TEST_JID,
            wildcard_resource=True,
        )

    def test_unregister_flattens_wildcard_resource_for_None(self):
        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            None,
            unittest.mock.sentinel.cb,
        )

        d.unregister_callback(
            unittest.mock.sentinel.type_,
            None,
            wildcard_resource=False,
        )

        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            None,
            unittest.mock.sentinel.cb,
        )

        d.unregister_callback(
            unittest.mock.sentinel.type_,
            None,
            wildcard_resource=True,
        )

    def test_unregister_raises_KeyError_if_unregistered(self):
        d = FooDispatcher()
        d.register_callback(
            unittest.mock.sentinel.type_,
            TEST_JID.bare(),
            unittest.mock.sentinel.cb,
            wildcard_resource=True,
        )

        with self.assertRaises(KeyError):
            d.unregister_callback(
                unittest.mock.sentinel.type_,
                TEST_JID.bare(),
                wildcard_resource=False,
            )

    def test_dispatch_converts_None_to_local_jid(self):
        self.d.register_callback(
            unittest.mock.sentinel.footype,
            TEST_LOCAL_JID,
            self.handlers.local,
        )

        stanza = FooStanza(None, unittest.mock.sentinel.footype)
        self.d._feed(stanza)
        self.assertCountEqual(
            self.handlers.mock_calls,
            [
                unittest.mock.call.local(stanza),
            ]
        )

    def test_dispatch_to_most_specific_type_fulljid(self):
        stanza = FooStanza(TEST_JID, unittest.mock.sentinel.type_)
        self.d._feed(stanza)
        self.assertCountEqual(
            self.handlers.mock_calls,
            [
                unittest.mock.call.type_fulljid_no_wildcard(stanza),
            ]
        )

    def test_dispatch_to_most_specific_type_fulljid_via_wildcard_to_bare(self):
        self.d.unregister_callback(
            unittest.mock.sentinel.type_,
            TEST_JID,
            wildcard_resource=False
        )

        stanza = FooStanza(TEST_JID, unittest.mock.sentinel.type_)
        self.d._feed(stanza)
        self.assertCountEqual(
            self.handlers.mock_calls,
            [
                unittest.mock.call.type_barejid_wildcard(stanza),
            ]
        )

    def test_dispatch_to_most_specific_type_fulljid_via_wildcard_to_none(self):
        self.d.unregister_callback(
            unittest.mock.sentinel.type_,
            TEST_JID,
            wildcard_resource=False
        )

        self.d.unregister_callback(
            unittest.mock.sentinel.type_,
            TEST_JID.bare(),
            wildcard_resource=True,
        )

        self.d.unregister_callback(
            None,
            TEST_JID.bare(),
            wildcard_resource=True,
        )

        self.d.unregister_callback(
            None,
            TEST_JID,
            wildcard_resource=False,
        )

        stanza = FooStanza(TEST_JID, unittest.mock.sentinel.type_)
        self.d._feed(stanza)
        self.assertCountEqual(
            self.handlers.mock_calls,
            [
                unittest.mock.call.type_wildcard(stanza),
            ]
        )

    def test_dispatch_to_most_specific_full_wildcard(self):
        stanza = FooStanza(TEST_JID.replace(localpart="fnord"),
                           unittest.mock.sentinel.othertype)
        self.d._feed(stanza)
        self.assertCountEqual(
            self.handlers.mock_calls,
            [
                unittest.mock.call.wildcard_wildcard(stanza),
            ]
        )

    def test_dispatch_to_most_specific_type_barejid_no_wildcard(self):
        stanza = FooStanza(TEST_JID.bare(), unittest.mock.sentinel.type_)
        self.d._feed(stanza)
        self.assertCountEqual(
            self.handlers.mock_calls,
            [
                unittest.mock.call.type_barejid_no_wildcard(stanza),
            ]
        )

    def test_dispatch_to_most_specific_mistype_fulljid(self):
        stanza = FooStanza(TEST_JID, unittest.mock.sentinel.othertype)
        self.d._feed(stanza)
        self.assertCountEqual(
            self.handlers.mock_calls,
            [
                unittest.mock.call.wildcard_fulljid_no_wildcard(stanza),
            ]
        )

    def test_dispatch_to_most_specific_mistype_fulljid_wildcard(self):
        self.d.unregister_callback(
            None,
            TEST_JID,
            wildcard_resource=False,
        )

        stanza = FooStanza(TEST_JID, unittest.mock.sentinel.othertype)
        self.d._feed(stanza)
        self.assertCountEqual(
            self.handlers.mock_calls,
            [
                unittest.mock.call.wildcard_barejid_wildcard(stanza),
            ]
        )

    def test_does_not_connect_to_on_message_received(self):
        self.assertFalse(
            aioxmpp.service.is_depsignal_handler(
                aioxmpp.stream.StanzaStream,
                "on_message_received",
                self.d._feed,
            )
        )

    def test_does_not_connect_to_on_presence_received(self):
        self.assertFalse(
            aioxmpp.service.is_depsignal_handler(
                aioxmpp.stream.StanzaStream,
                "on_presence_received",
                self.d._feed,
            )
        )

    def test_handler_context_is_context_manager(self):
        cm = self.d.handler_context(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
            unittest.mock.sentinel.cb,
            wildcard_resource=unittest.mock.sentinel.wildcard_resource,
        )

        self.assertTrue(hasattr(cm, "__enter__"))
        self.assertTrue(hasattr(cm, "__exit__"))

    def test_handler_context_enter_registers_callback(self):
        cm = self.d.handler_context(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
            unittest.mock.sentinel.cb,
            wildcard_resource=unittest.mock.sentinel.wildcard_resource,
        )

        # we need to mock this too, but we don’t want to test __exit__ here
        with unittest.mock.patch.object(
                self.d,
                "unregister_callback") as unregister_callback:
            with unittest.mock.patch.object(
                    self.d,
                    "register_callback") as register_callback:
                cm.__enter__()

            register_callback.assert_called_once_with(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                unittest.mock.sentinel.cb,
                wildcard_resource=unittest.mock.sentinel.wildcard_resource,
            )

            unregister_callback.assert_not_called()

            cm.__exit__(None, None, None)

    def test_handler_context_exit_unregisters_callback(self):
        cm = self.d.handler_context(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
            unittest.mock.sentinel.cb,
            wildcard_resource=unittest.mock.sentinel.wildcard_resource,
        )

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch.object(
                    self.d,
                    "register_callback")
            )
            unregister_callback = stack.enter_context(
                unittest.mock.patch.object(
                    self.d,
                    "unregister_callback")
            )

            cm.__enter__()
            cm.__exit__(None, None, None)

        unregister_callback.assert_called_once_with(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
            wildcard_resource=unittest.mock.sentinel.wildcard_resource,
        )

    def test_handler_context_exit_unregisters_and_does_not_swallow(self):
        cm = self.d.handler_context(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
            unittest.mock.sentinel.cb,
            wildcard_resource=unittest.mock.sentinel.wildcard_resource,
        )

        class FooException(Exception):
            pass

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch.object(
                    self.d,
                    "register_callback")
            )
            unregister_callback = stack.enter_context(
                unittest.mock.patch.object(
                    self.d,
                    "unregister_callback")
            )
            stack.enter_context(
                self.assertRaises(FooException)
            )
            stack.enter_context(cm)
            raise FooException()

        unregister_callback.assert_called_once_with(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
            wildcard_resource=unittest.mock.sentinel.wildcard_resource,
        )


class TestSimpleMessageDispatcher(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.d = dispatcher.SimpleMessageDispatcher(self.cc)

    def tearDown(self):
        del self.d
        del self.cc

    def test_is_service(self):
        self.assertTrue(issubclass(
            dispatcher.SimpleMessageDispatcher,
            aioxmpp.service.Service,
        ))

    def test_is_SimpleStanzaDispatcher(self):
        self.assertTrue(issubclass(
            dispatcher.SimpleMessageDispatcher,
            dispatcher.SimpleStanzaDispatcher,
        ))

    def test_local_jid_uses_local_jid_from_client(self):
        self.assertEqual(
            self.d.local_jid,
            self.cc.local_jid,
        )

    def test_connects_to_on_message_received(self):
        self.assertTrue(
            aioxmpp.service.is_depsignal_handler(
                aioxmpp.stream.StanzaStream,
                "on_message_received",
                self.d._feed,
            )
        )


class TestSimplePresenceDispatcher(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.d = dispatcher.SimplePresenceDispatcher(self.cc)

    def tearDown(self):
        del self.d
        del self.cc

    def test_is_service(self):
        self.assertTrue(issubclass(
            dispatcher.SimplePresenceDispatcher,
            aioxmpp.service.Service,
        ))

    def test_is_SimpleStanzaDispatcher(self):
        self.assertTrue(issubclass(
            dispatcher.SimplePresenceDispatcher,
            dispatcher.SimpleStanzaDispatcher,
        ))

    def test_local_jid_uses_local_jid_from_client(self):
        self.assertEqual(
            self.d.local_jid,
            self.cc.local_jid,
        )

    def test_connects_to_on_presence_received(self):
        self.assertTrue(
            aioxmpp.service.is_depsignal_handler(
                aioxmpp.stream.StanzaStream,
                "on_presence_received",
                self.d._feed,
            )
        )


class Test_apply_message_handler(unittest.TestCase):
    def test_uses_SimpleMessageDispatcher(self):
        instance = unittest.mock.MagicMock()
        dependency = unittest.mock.Mock()
        instance.dependencies.__getitem__.return_value = dependency

        result = dispatcher._apply_message_handler(
            instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
        )

        instance.dependencies.__getitem__.assert_called_once_with(
            dispatcher.SimpleMessageDispatcher,
        )

        dependency.handler_context.assert_called_with(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
            unittest.mock.sentinel.func,
        )

        self.assertEqual(
            result,
            dependency.handler_context(),
        )


class Test_apply_presence_handler(unittest.TestCase):
    def test_uses_SimplePresenceDispatcher(self):
        instance = unittest.mock.MagicMock()
        dependency = unittest.mock.Mock()
        instance.dependencies.__getitem__.return_value = dependency

        result = dispatcher._apply_presence_handler(
            instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
        )

        instance.dependencies.__getitem__.assert_called_once_with(
            dispatcher.SimplePresenceDispatcher,
        )

        dependency.handler_context.assert_called_with(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_,
            unittest.mock.sentinel.func,
        )

        self.assertEqual(
            result,
            dependency.handler_context(),
        )


class Testmessage_handler(unittest.TestCase):
    def setUp(self):
        self.decorator = dispatcher.message_handler(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_
        )

    def tearDown(self):
        del self.decorator

    def test_works_as_decorator(self):
        def cb():
            pass

        self.assertIs(
            cb,
            self.decorator(cb),
        )

    def test_adds_magic_attribute(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertTrue(hasattr(cb, "_aioxmpp_service_handlers"))

    def test_adds__apply_message_handler_entry(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertIn(
            aioxmpp.service.HandlerSpec(
                (dispatcher._apply_message_handler,
                 (unittest.mock.sentinel.type_,
                  unittest.mock.sentinel.from_)),
                is_unique=True,
                require_deps=(
                    dispatcher.SimpleMessageDispatcher,
                ),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_stacks_with_other_effects(self):
        def cb():
            pass

        cb._aioxmpp_service_handlers = {"foo"}

        self.decorator(cb)

        self.assertIn(
            aioxmpp.service.HandlerSpec(
                (dispatcher._apply_message_handler,
                 (unittest.mock.sentinel.type_,
                  unittest.mock.sentinel.from_)),
                is_unique=True,
                require_deps=(
                    dispatcher.SimpleMessageDispatcher,
                ),
            ),
            cb._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            cb._aioxmpp_service_handlers,
        )

    def test_requires_non_coroutine(self):
        with unittest.mock.patch(
                "asyncio.iscoroutinefunction") as iscoroutinefunction:
            iscoroutinefunction.return_value = True

            with self.assertRaisesRegex(
                    TypeError,
                    "must not be a coroutine function"):
                self.decorator(unittest.mock.sentinel.cb)

        iscoroutinefunction.assert_called_with(
            unittest.mock.sentinel.cb,
        )

    def test_works_with_is_message_handler(self):
        def cb():
            pass

        self.assertFalse(
            dispatcher.is_message_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                cb,
            )
        )

        self.decorator(cb)

        self.assertTrue(
            dispatcher.is_message_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                cb,
            )
        )


class Testpresence_handler(unittest.TestCase):
    def setUp(self):
        self.decorator = dispatcher.presence_handler(
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.from_
        )

    def tearDown(self):
        del self.decorator

    def test_works_as_decorator(self):
        def cb():
            pass

        self.assertIs(
            cb,
            self.decorator(cb),
        )

    def test_adds_magic_attribute(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertTrue(hasattr(cb, "_aioxmpp_service_handlers"))

    def test_adds__apply_presence_handler_entry(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertIn(
            aioxmpp.service.HandlerSpec(
                (dispatcher._apply_presence_handler,
                 (unittest.mock.sentinel.type_,
                  unittest.mock.sentinel.from_)),
                is_unique=True,
                require_deps=(
                    dispatcher.SimplePresenceDispatcher,
                ),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_stacks_with_other_effects(self):
        def cb():
            pass

        cb._aioxmpp_service_handlers = {"foo"}

        self.decorator(cb)

        self.assertIn(
            aioxmpp.service.HandlerSpec(
                (dispatcher._apply_presence_handler,
                 (unittest.mock.sentinel.type_,
                  unittest.mock.sentinel.from_)),
                is_unique=True,
                require_deps=(
                    dispatcher.SimplePresenceDispatcher,
                ),
            ),
            cb._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            cb._aioxmpp_service_handlers,
        )

    def test_requires_non_coroutine(self):
        with unittest.mock.patch(
                "asyncio.iscoroutinefunction") as iscoroutinefunction:
            iscoroutinefunction.return_value = True

            with self.assertRaisesRegex(
                    TypeError,
                    "must not be a coroutine function"):
                self.decorator(unittest.mock.sentinel.cb)

        iscoroutinefunction.assert_called_with(
            unittest.mock.sentinel.cb,
        )

    def test_works_with_is_presence_handler(self):
        def cb():
            pass

        self.assertFalse(
            dispatcher.is_presence_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                cb,
            )
        )

        self.decorator(cb)

        self.assertTrue(
            dispatcher.is_presence_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                cb,
            )
        )


class Testis_message_handler(unittest.TestCase):
    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            dispatcher.is_message_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                object()
            )
        )

    def test_return_true_if_token_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            aioxmpp.service.HandlerSpec(
                (dispatcher._apply_message_handler,
                    (unittest.mock.sentinel.type_,
                     unittest.mock.sentinel.from_)),
                require_deps=(
                    dispatcher.SimpleMessageDispatcher,
                )
            ),
        ]

        self.assertTrue(
            dispatcher.is_message_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                m
            )
        )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            aioxmpp.service.HandlerSpec(
                (dispatcher._apply_message_handler,
                 (unittest.mock.sentinel.type2,
                     unittest.mock.sentinel.from2)),
                require_deps=(
                    dispatcher.SimpleMessageDispatcher,
                )
            )
        ]

        self.assertFalse(
            dispatcher.is_message_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                m
            )
        )


class Testis_presence_handler(unittest.TestCase):
    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            dispatcher.is_presence_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                object()
            )
        )

    def test_return_true_if_token_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            aioxmpp.service.HandlerSpec(
                (dispatcher._apply_presence_handler,
                 (unittest.mock.sentinel.type_,
                  unittest.mock.sentinel.from_)),
                require_deps=(
                    dispatcher.SimplePresenceDispatcher,
                )
            )
        ]

        self.assertTrue(
            dispatcher.is_presence_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                m
            )
        )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            aioxmpp.service.HandlerSpec(
                (dispatcher._apply_presence_handler,
                 (unittest.mock.sentinel.type2,
                  unittest.mock.sentinel.from2)),
                require_deps=(
                    dispatcher.SimplePresenceDispatcher,
                )
            )
        ]

        self.assertFalse(
            dispatcher.is_presence_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.from_,
                m
            )
        )



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.forms` --- Data Forms support (:xep:`4`)
#######################################################

This subpackage contains tools to deal with :xep:`4` Data Forms. Data Forms is
a pervasive and highly flexible protocol used in XMPP. It allows for
machine-readable (and processable) forms as well as tables of data. This
flexibility comes unfortunately at the price of complexity. This subpackage
attempts to take some of the load of processing Data Forms off the application
developer.

Cheat Sheet:

* The :class:`Form` class exists for use cases where automated processing of
  Data Forms is supposed to happen. The
  :ref:`api-aioxmpp.forms-declarative-style` allow convenient access to and
  manipulation of form data from within code.

* Direct use of the :class:`Data` XSO is advisable if you want to present forms
  or data to sentient beings: even though :class:`Form` is more convenient for
  machine-to-machine use, using the :class:`Data` sent by the peer easily
  allows showing the user *all* fields supported by the peer.

* For machine-processed tables, there is no tooling (yet).

.. versionadded:: 0.7

   Even though the :mod:`aioxmpp.forms` module existed pre-0.7, it has not been
   documented and was thus not part of the public API.

.. note::

   The authors are not entirely happy with the API at some points.
   Specifically, at some places where mutable data structures are used, the
   mutation of these data structures may have unexpected side effects. This may
   be rectified in a future release by replacing these data structures with
   their appropriate immutable equivalents.

   These locations are marked accordingly.

Attributes added to stanzas
===========================

:mod:`aioxmpp.forms` adds the following attributes to stanzas:

.. attribute:: aioxmpp.Message.xep0004_data

   A sequence of :class:`Data` instances. This is used for example by the
   :mod:`~.muc` implementation (:xep:`45`).

   .. versionadded:: 0.8


.. _api-aioxmpp.forms-declarative-style:

Declarative-style Forms
=======================

Base class
----------

.. autoclass:: Form

Fields
------

Text fields
~~~~~~~~~~~

.. autoclass:: TextSingle(var, type_=xso.String(), *[, default=None][, required=False][, desc=None][, label=None])

.. autoclass:: TextPrivate(var, type_=xso.String(), *[, default=None][, required=False][, desc=None][, label=None])

.. autoclass:: TextMulti(var, type_=xso.String(), *[, default=()][, required=False][, desc=None][, label=None])

JID fields
~~~~~~~~~~

.. autoclass:: JIDSingle(var, *[, default=None][, required=False][, desc=None][, label=None])

.. autoclass:: JIDMulti(var, *[, default=()][, required=False][, desc=None][, label=None])

Selection fields
~~~~~~~~~~~~~~~~

.. autoclass:: ListSingle(var, type_=xso.String(), *[, default=None][, options=[]][, required=False][, desc=None][, label=None])

.. autoclass:: ListMulti(var, type_=xso.String(), *[, default=frozenset()][, options=[]][, required=False][, desc=None][, label=None])

Other fields
~~~~~~~~~~~~

.. autoclass:: Boolean(var, *[, default=False][, required=False][, desc=None][, label=None])

Abstract base classes
~~~~~~~~~~~~~~~~~~~~~

.. currentmodule:: aioxmpp.forms.fields

.. autoclass:: AbstractField

.. autoclass:: AbstractChoiceField(var, type_=xso.String(), *[, options=[]][, required=False][, desc=None][, label=None])

.. currentmodule:: aioxmpp.forms

.. _api-aioxmpp.forms-bound-fields:

Bound fields
============

Bound fields are objects which are returned when the descriptor attribute is
accessed on a form instance. It holds the value of the field, as well as
overrides for the default (specified on the descriptors themselves) values for
certain attributes (such as :attr:`~.AbstractField.desc`).

For the different field types, there are different classes of bound fields,
which are documented below.

.. currentmodule:: aioxmpp.forms.fields

.. autoclass:: BoundField

.. autoclass:: BoundSingleValueField

.. autoclass:: BoundMultiValueField

.. autoclass:: BoundOptionsField

.. autoclass:: BoundSelectField

.. autoclass:: BoundMultiSelectField

.. currentmodule:: aioxmpp.forms

XSOs
====

.. autoclass:: Data

.. autoclass:: DataType

.. autoclass:: Field

.. autoclass:: FieldType

Report and table support
------------------------

.. autoclass:: Reported

.. autoclass:: Item

"""


from . import xso

from .xso import (  # NOQA
    Data,
    DataType,
    Field,
    FieldType,
    Reported,
    Item,
)

from .fields import (  # NOQA
    Boolean,
    ListSingle,
    ListMulti,
    JIDSingle,
    JIDMulti,
    TextSingle,
    TextMulti,
    TextPrivate,
)

from .form import (  # NOQA
    Form,
)



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import functools

import aioxmpp.callbacks
import aioxmpp.service


class ConversationService(aioxmpp.service.Service):
    """
    Central place where all :class:`.im.conversation.AbstractConversation`
    subclass instances are collected.

    It provides discoverability of all existing conversations (in no particular
    order) and signals on addition and removal of active conversations. This is
    useful for front ends to track conversations globally without needing to
    know about the specific conversation providers.

    .. signal:: on_conversation_added(conversation)

       A new conversation has been added.

       :param conversation: The conversation which was added.
       :type conversation: :class:`~.im.conversation.AbstractConversation`

       This signal is fired when a new conversation is added by a
       :term:`Conversation Implementation`.

       .. note::

          If you are looking for a "on_conversation_removed" event or similar,
          there is none. You should use the
          :meth:`.AbstractConversation.on_exit` event of the `conversation`.

    .. signal:: on_message(conversation,
                           *args, **kwargs)

        Emits whenever any active conversation emits its
        :meth:`~.im.Conversation.on_message` event. The arguments are forwarded
        1:1, with the :class:`~.im.AbstractConversation` instance pre-pended to
        the argument list.

    .. autoattribute:: conversations

    .. automethod:: get_conversation

    For :term:`Conversation Implementations <Conversation Implementation>`, the
    following methods are intended; they should not be used by applications.

    .. automethod:: _add_conversation

    """

    on_conversation_added = aioxmpp.callbacks.Signal()
    on_message = aioxmpp.callbacks.Signal()

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        self._conversation_meta = {}
        self._conversation_map = {}

    @property
    def conversations(self):
        """
        Return an iterable of conversations in which the local client is
        participating.
        """
        return self._conversation_meta.keys()

    def _remove_conversation(self, conv):
        del self._conversation_map[conv.jid]
        tokens, = self._conversation_meta.pop(conv)
        for signal, token in tokens:
            signal.disconnect(token)

    def _handle_conversation_exit(self, conv, *args, **kwargs):
        self._remove_conversation(conv)
        return False

    def _add_conversation(self, conversation):
        """
        Add the conversation and fire the :meth:`on_conversation_added` event.

        :param conversation: The conversation object to add.
        :type conversation: :class:`~.AbstractConversation`

        The conversation is added to the internal list of conversations which
        can be queried at :attr:`conversations`. The
        :meth:`on_conversation_added` event is fired.

        In addition, the :class:`ConversationService` subscribes to the
        :meth:`~.AbstractConversation.on_exit` event to remove the conversation
        from the list automatically. There is no need to remove a conversation
        from the list explicitly.
        """
        handler = functools.partial(
            self._handle_conversation_exit,
            conversation
        )
        tokens = []

        def linked_token(signal, handler):
            return signal, signal.connect(handler)

        tokens.append(linked_token(conversation.on_exit, handler))
        tokens.append(linked_token(conversation.on_failure, handler))
        tokens.append(linked_token(conversation.on_message, functools.partial(
            self.on_message,
            conversation,
        )))

        self._conversation_meta[conversation] = (
            tokens,
        )
        self._conversation_map[conversation.jid] = conversation
        self.on_conversation_added(conversation)

    def get_conversation(self, conversation_address):
        """
        Return the :class:`.im.AbstractConversation` for a given JID.

        :raises KeyError: if there is currently no matching conversation
        """
        return self._conversation_map[conversation_address]



########################################################################
# File name: echo_bot.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp

from framework import Example, exec_example


class EchoBot(Example):
    def message_received(self, msg):
        if msg.type_ != aioxmpp.MessageType.CHAT:
            return

        if not msg.body:
            # do not reflect anything without a body
            return

        # we could also use reply = msg.make_reply() instead
        reply = aioxmpp.Message(
            type_=msg.type_,
            to=msg.from_,
        )

        # make_reply() would not set the body though
        reply.body.update(msg.body)

        self.client.enqueue(reply)

    @asyncio.coroutine
    def run_simple_example(self):
        stop_event = self.make_sigint_event()

        self.client.stream.register_message_callback(
            aioxmpp.MessageType.CHAT,
            None,
            self.message_received,
        )
        print("echoing... (press Ctrl-C or send SIGTERM to stop)")
        yield from stop_event.wait()

if __name__ == "__main__":
    exec_example(EchoBot())



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.disco
import aioxmpp.shim.service as shim_service

from aioxmpp.testutils import (
    make_connected_client,
    CoroutineMock,
    run_coroutine,
)

TEST_FROM = aioxmpp.structs.JID.fromstr("foo@bar.example/baz")


class TestService(unittest.TestCase):
    def setUp(self):
        self.disco = unittest.mock.Mock()
        self.disco.query_info = CoroutineMock()
        self.disco.query_info.side_effect = AssertionError
        self.disco.query_items = CoroutineMock()
        self.disco.query_items.side_effect = AssertionError
        self.node = unittest.mock.Mock()
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_FROM

        with unittest.mock.patch("aioxmpp.disco.StaticNode") as Node:
            Node.return_value = self.node
            self.s = shim_service.SHIMService(self.cc, dependencies={
                aioxmpp.DiscoServer: self.disco
            })

        self.disco.mock_calls.clear()
        self.cc.mock_calls.clear()

    def tearDown(self):
        del self.s
        del self.cc
        del self.disco

    def test_orders_before_disco_service(self):
        self.assertLess(
            aioxmpp.DiscoServer,
            shim_service.SHIMService,
        )

    def test_init(self):
        self.disco = unittest.mock.Mock()
        self.disco.query_info = CoroutineMock()
        self.disco.query_info.side_effect = AssertionError
        self.disco.query_items = CoroutineMock()
        self.disco.query_items.side_effect = AssertionError
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_FROM

        with unittest.mock.patch("aioxmpp.disco.StaticNode") as Node:
            self.s = shim_service.SHIMService(self.cc, dependencies={
                aioxmpp.DiscoServer: self.disco,
            })

        self.disco.register_feature.assert_called_with(
            "http://jabber.org/protocol/shim"
        )

        self.disco.mount_node.assert_called_with(
            "http://jabber.org/protocol/shim",
            Node()
        )

    def test_shutdown(self):
        run_coroutine(self.s.shutdown())
        self.disco.unregister_feature.assert_called_with(
            "http://jabber.org/protocol/shim"
        )

        self.disco.unmount_node.assert_called_with(
            "http://jabber.org/protocol/shim",
        )

    def test_register_header(self):
        self.s.register_header(
            "Foo"
        )

        self.node.register_feature.assert_called_with(
            "http://jabber.org/protocol/shim#Foo"
        )

    def test_unregister_header(self):
        self.s.unregister_header(
            "Foo"
        )

        self.node.unregister_feature.assert_called_with(
            "http://jabber.org/protocol/shim#Foo"
        )



########################################################################
# File name: types.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`aioxmpp.xso.types` --- Types specifications for use with :mod:`aioxmpp.xso.model`
#######################################################################################

See :mod:`aioxmpp.xso` for documentation.

"""

import abc
import array
import base64
import binascii
import decimal
import ipaddress
import numbers
import re
import unicodedata
import warnings

import pytz

from datetime import datetime, timedelta, date, time

from .. import structs, i18n


class Unknown:
    """
    A wrapper for an unknown enumeration value.

    :param value: The raw value of the "enumeration" "member".
    :type value: arbitrary

    Instances of this class may be emitted from and accepted by
    :class:`EnumCDataType` and :class:`EnumElementType`, see the documentation
    there for details.

    :class:`Unknown` instances compare equal when they hold an equal value.
    :class:`Unknown` objects are hashable if their values are hashable. The
    value they refer to cannot be changed during the lifetime of an
    :class:`Unknown` object.
    """

    def __init__(self, value):
        super().__init__()
        self.__value = value

    @property
    def value(self):
        return self.__value

    def __hash__(self):
        return hash(self.__value)

    def __eq__(self, other):
        try:
            return self.__value == other.__value
        except AttributeError:
            return NotImplemented

    def __repr__(self):
        return "<Unknown: {!r}>".format(
            self.__value
        )


class AbstractCDataType(metaclass=abc.ABCMeta):
    """
    Subclasses of this class describe character data types.

    They are used to convert python values from (:meth:`parse`) and to
    (:meth:`format`) XML character data as well as enforce basic type
    restrictions (:meth:`coerce`) when values are assigned to descriptors
    using this type.

    This type can be used by the character data descriptors, like :class:`Attr`
    and :class:`Text`.

    .. automethod:: coerce

    .. automethod:: parse

    .. automethod:: format
    """

    def coerce(self, v):
        """
        Force the given value `v` to be of the type represented by this
        :class:`AbstractCDataType`.

        :meth:`coerce` is called when user code assigns values to descriptors
        which use the type; it is notably not called when values are extracted
        from SAX events, as these go through :meth:`parse` and that is expected
        to return correctly typed values.

        If `v` cannot be sensibly coerced, :class:`TypeError` is raised (in
        some rare occasions, :class:`ValueError` may be ok too).

        Return a coerced version of `v` or `v` itself if it matches the
        required type.

        .. note::

           For the sake of usability, coercion should only take place rarely;
           in most of the cases, throwing :class:`TypeError` is the preferred
           method.

           Otherwise, a user might be surprised why the :class:`int` they
           assigned to an attribute suddenly became a :class:`str`.

        """
        return v

    @abc.abstractmethod
    def parse(self, v):
        """
        Convert the given string `v` into a value of the appropriate type this
        class implements and return the result.

        If conversion fails, :class:`ValueError` is raised.

        The result of :meth:`parse` must pass through :meth:`coerce` unchanged.
        """

    def format(self, v):
        """
        Convert the value `v` of the type this class implements to a str.

        This conversion does not fail.

        The returned value can be passed to :meth:`parse` to obtain `v`.
        """
        return str(v)


class AbstractElementType(metaclass=abc.ABCMeta):
    """
    Subclasses of this class describe XML subtree types.

    They are used to convert python values from (:meth:`unpack`) and to
    (:meth:`pack`) XML subtrees represented as :class:`XSO` instances as well
    as enforce basic type restrictions (:meth:`coerce`) when values are
    assigned to descriptors using this type.

    This type can be used by the element descriptors, like
    :class:`ChildValueList` and :class:`ChildValueMap`.

    .. automethod:: get_xso_types

    .. automethod:: coerce

    .. automethod:: unpack

    .. automethod:: pack
    """

    @abc.abstractmethod
    def get_xso_types(self):
        """
        Return the :class:`XSO` subclasses supported by this type.

        :rtype: :class:`~collections.Iterable` of :class:`XMLStreamClass`
        :return: The :class:`XSO` subclasses which can be passed to
            :meth:`unpack`.
        """

    @abc.abstractmethod
    def unpack(self, obj):
        """
        Convert a :class:`XSO` instance to another object, usually a scalar
        value or a tuple.

        :param obj: The object to unpack.
        :type obj: One of the types returned by :meth:`get_xso_types`.
        :raises ValueError: if the conversaion fails.
        :return: The unpacked value.

        Think of unpack like a high-level :func:`struct.unpack`: it converts
        wire-format data (XML subtrees represented as :class:`XSO` instances)
        to python values.
        """

    @abc.abstractmethod
    def pack(self, v):
        """
        Convert the value `v` of the type this class implements to an
        :class:`XSO` instance.

        :param v: Value to pack
        :type v: as returned by :meth:`unpack`
        :rtype: One of the types returned by :meth:`get_xso_types`.
        :return: The packed value.

        The returned value can be passed through :meth:`unpack` to obtain a
        value equal to `v`.

        Think of pack like a high-level :func:`struct.pack`: it converts
        python values to wire-format (XML subtrees represented as :class:`XSO`
        instances).
        """

    def coerce(self, v):
        """
        Force the given value `v` to be compatible to :meth:`pack`.

        :meth:`coerce` is called when user code assigns
        values to descriptors which use the type; it is notably not called when
        values are extracted from SAX events, as these go through
        :meth:`unpack` and that is expected to return correctly typed values.

        If `v` cannot be sensibly coerced, :class:`TypeError` is raised (in
        some rare occasions, :class:`ValueError` may be ok too).

        Return a coerced version of `v` or `v` itself if it matches the
        required type.

        .. note::

           For the sake of usability, coercion should only take place rarely;
           in most of the cases, throwing :class:`TypeError` is the preferred
           method.

           Otherwise, a user might be surprised why the :class:`int` they
           assigned to an attribute suddenly became a :class:`str`.

        """
        return v


class String(AbstractCDataType):
    """
    Interpret the input value as string.

    Optionally, a stringprep function `prepfunc` can be applied on the
    string. A stringprep function must take the string and prepare it
    accordingly; if it is invalid input, it must raise
    :class:`ValueError`. Otherwise, it shall return the prepared string.

    If no `prepfunc` is given, this type is the identity operation.
    """

    def __init__(self, prepfunc=None):
        super().__init__()
        self.prepfunc = prepfunc

    def coerce(self, v):
        if not isinstance(v, str):
            raise TypeError("must be a str object")
        if self.prepfunc is not None:
            return self.prepfunc(v)
        return v

    def parse(self, v):
        if self.prepfunc is not None:
            return self.prepfunc(v)
        return v


class Integer(AbstractCDataType):
    """
    Parse the value as base-10 integer and return the result as :class:`int`.
    """

    def coerce(self, v):
        if not isinstance(v, numbers.Integral):
            raise TypeError("must be integral number")
        return int(v)

    def parse(self, v):
        return int(v)


class Float(AbstractCDataType):
    """
    Parse the value as decimal float and return the result as :class:`float`.
    """

    def coerce(self, v):
        if not isinstance(v, (numbers.Real, decimal.Decimal)):
            raise TypeError("must be real number")
        return float(v)

    def parse(self, v):
        return float(v)


class Bool(AbstractCDataType):
    """
    Parse the value as boolean:

    * ``"true"`` and ``"1"`` are taken as :data:`True`,
    * ``"false"`` and ``"0"`` are taken as :data:`False`,
    * everything else results in a :class:`ValueError` exception.

    """

    def coerce(self, v):
        return bool(v)

    def parse(self, v):
        v = v.strip()
        if v in ["true", "1"]:
            return True
        elif v in ["false", "0"]:
            return False
        else:
            raise ValueError("not a boolean value")

    def format(self, v):
        if v:
            return "true"
        else:
            return "false"


class DateTime(AbstractCDataType):
    """
    Parse the value as ISO datetime, possibly including microseconds and
    timezone information.

    Timezones are handled as constant offsets from UTC, and are converted to
    UTC before the :class:`~datetime.datetime` object is returned (which is
    correctly tagged with UTC tzinfo). Values without timezone specification
    are not tagged.

    If `legacy` is true, the formatted dates use the legacy date/time format
    (``CCYYMMDDThh:mm:ss``), as used for example in :xep:`0082` or :xep:`0009`
    (whereas in the latter it is not legacy, but defined by XML RPC). In any
    case, parsing of the legacy format is transparently supported. Timestamps
    in the legacy format are assumed to be in UTC, and datetime objects are
    converted to UTC before emitting the legacy format. The timezone designator
    is never emitted with the legacy format, and ignored if given.

    This class makes use of :mod:`pytz`.

    .. versionadded:: 0.5

       The `legacy` argument was added.

    """

    tzextract = re.compile("((Z)|([+-][0-9]{2}):([0-9]{2}))$")

    def __init__(self, *, legacy=False):
        super().__init__()
        self.legacy = legacy

    def coerce(self, v):
        if not isinstance(v, datetime):
            raise TypeError("must be a datetime object")
        return v

    def parse(self, v):
        v = v.strip()
        m = self.tzextract.search(v)
        if m:
            _, utc, hour_offset, minute_offset = m.groups()
            if utc:
                hour_offset = 0
                minute_offset = 0
            else:
                hour_offset = int(hour_offset)
                minute_offset = int(minute_offset)
            tzinfo = pytz.utc
            offset = timedelta(minutes=minute_offset + 60 * hour_offset)
            v = v[:m.start()]
        else:
            tzinfo = None
            offset = timedelta(0)

        try:
            dt = datetime.strptime(v, "%Y-%m-%dT%H:%M:%S.%f")
        except ValueError:
            try:
                dt = datetime.strptime(v, "%Y-%m-%dT%H:%M:%S")
            except ValueError:
                dt = datetime.strptime(v, "%Y%m%dT%H:%M:%S")
                tzinfo = pytz.utc
                offset = timedelta(0)

        return dt.replace(tzinfo=tzinfo) - offset

    def format(self, v):
        if v.tzinfo:
            v = pytz.utc.normalize(v)
        if self.legacy:
            return v.strftime("%Y%m%dT%H:%M:%S")

        result = v.strftime("%Y-%m-%dT%H:%M:%S")
        if v.microsecond:
            result += ".{:06d}".format(v.microsecond).rstrip("0")
        if v.tzinfo:
            result += "Z"
        return result


class Date(AbstractCDataType):
    """
    Implement the Date type from :xep:`0082`.

    Values must have the :class:`date` type, :class:`datetime` is forbidden to
    avoid silent loss of information.

    .. versionadded:: 0.5
    """

    def parse(self, s):
        return datetime.strptime(s, "%Y-%m-%d").date()

    def coerce(self, v):
        if not isinstance(v, date) or isinstance(v, datetime):
            raise TypeError("must be a date object")
        return v


class Time(AbstractCDataType):
    """
    Implement the Time type from :xep:`0082`.

    Values must have the :class:`time` type, :class:`datetime` is forbidden to
    avoid silent loss of information. Assignment of :class:`time` values in
    time zones which are not UTC is not allowed either. The reason is that the
    translation to UTC on formatting is not properly defined without an
    accompanying date (think daylight saving time transitions, redefinitions of
    time zones, …).

    .. versionadded:: 0.5
    """

    def parse(self, v):
        v = v.strip()
        m = DateTime.tzextract.search(v)
        if m:
            _, utc, hour_offset, minute_offset = m.groups()
            if utc:
                hour_offset = 0
                minute_offset = 0
            else:
                hour_offset = int(hour_offset)
                minute_offset = int(minute_offset)
            tzinfo = pytz.utc
            offset = timedelta(minutes=minute_offset + 60 * hour_offset)
            v = v[:m.start()]
        else:
            tzinfo = None
            offset = timedelta(0)

        try:
            dt = datetime.strptime(v, "%H:%M:%S.%f")
        except ValueError:
            dt = datetime.strptime(v, "%H:%M:%S")

        return (dt.replace(tzinfo=tzinfo) - offset).timetz()

    def format(self, v):
        if v.tzinfo:
            v = pytz.utc.normalize(v)

        result = v.strftime("%H:%M:%S")
        if v.microsecond:
            result += ".{:06d}".format(v.microsecond).rstrip("0")
        if v.tzinfo:
            result += "Z"
        return result

    def coerce(self, t):
        if not isinstance(t, time):
            raise TypeError("must be a time object")
        if t.tzinfo is None:
            return t
        if t.tzinfo == pytz.utc:
            return t
        raise ValueError("time must have UTC timezone or none at all")


class _BinaryType(AbstractCDataType):
    """
    Implements pointful coercion for binary types.
    """

    def coerce(self, v):
        if isinstance(v, bytes):
            return v
        elif isinstance(v, (bytearray, array.array)):
            return bytes(v)
        raise TypeError("must be convertible to bytes")


class Base64Binary(_BinaryType):
    """
    Parse the value as base64 and return the :class:`bytes` object obtained
    from decoding.

    If `empty_as_equal` is :data:`True`, an empty value is represented using a
    single equal sign. This is used in the SASL protocol.
    """

    def __init__(self, *, empty_as_equal=False):
        super().__init__()
        self._empty_as_equal = empty_as_equal

    def parse(self, v):
        return base64.b64decode(v)

    def format(self, v):
        if self._empty_as_equal and not v:
            return "="
        return base64.b64encode(v).decode("ascii")


class HexBinary(_BinaryType):
    """
    Parse the value as hexadecimal blob and return the :class:`bytes` object
    obtained from decoding.
    """

    def parse(self, v):
        return binascii.a2b_hex(v)

    def format(self, v):
        return binascii.b2a_hex(v).decode("ascii")


class JID(AbstractCDataType):
    """
    Parse the value as Jabber ID using :meth:`~aioxmpp.JID.fromstr` and
    return the :class:`aioxmpp.JID` object.

    `strict` is passed to :meth:`~aioxmpp.JID.fromstr` and defaults to
    false. See the :meth:`~aioxmpp.JID.fromstr` method for a rationale
    and consider that :meth:`parse` is only called for input coming from the
    outside.
    """

    def __init__(self, *, strict=False):
        super().__init__()
        self.strict = strict

    def coerce(self, v):
        if not isinstance(v, structs.JID):
            raise TypeError("{} object {!r} is not a JID".format(
                type(v), v))

        return v

    def parse(self, v):
        return structs.JID.fromstr(v, strict=self.strict)


class ConnectionLocation(AbstractCDataType):
    """
    Parse the value as a host-port pair, as for example used for Stream
    Management reconnection location advisories.
    """

    def coerce(self, v):
        if not isinstance(v, tuple):
            raise TypeError("2-tuple required for ConnectionLocation")
        if len(v) != 2:
            raise TypeError("2-tuple required for ConnectionLocation")

        addr, port = v

        if not isinstance(port, numbers.Integral):
            raise TypeError("port number must be integral number")
        port = int(port)

        if not (0 <= port <= 65535):
            raise ValueError("port number {} out of range".format(port))

        try:
            addr = ipaddress.IPv4Address(addr)
        except ValueError:
            try:
                addr = ipaddress.IPv6Address(addr)
            except ValueError:
                pass

        return (addr, port)

    def parse(self, v):
        v = v.strip()
        addr, _, port = v.rpartition(":")
        if not _:
            raise ValueError("missing colon in connection location")
        port = int(port)

        if addr.startswith("[") and addr.endswith("]"):
            addr = ipaddress.IPv6Address(addr[1:-1])

        try:
            addr = ipaddress.IPv4Address(addr)
        except ValueError:
            pass

        return self.coerce((addr, port))

    def format(self, v):
        if isinstance(v[0], ipaddress.IPv6Address):
            return "[{}]:{}".format(*v)
        return ":".join(map(str, v))


class LanguageTag(AbstractCDataType):
    """
    Parses the value as Language Tag using
    :meth:`~.structs.LanguageTag.fromstr`.

    Type coercion requires that any value assigned to a descriptor using this
    type is an instance of :class:`~.structs.LanguageTag`.
    """

    def parse(self, v):
        return structs.LanguageTag.fromstr(v)

    def coerce(self, v):
        if not isinstance(v, structs.LanguageTag):
            raise TypeError("{!r} is not a LanguageTag", v)
        return v


class TextChildMap(AbstractElementType):
    """
    A type for use with :class:`.xso.ChildValueMap` and descendants of
    :class:`.xso.AbstractTextChild`.

    This type performs the packing and unpacking of language-text-pairs to and
    from the `xso_type`. `xso_type` must have an interface compatible with
    :class:`.xso.AbstractTextChild`, which means that it must have the language
    and text at :attr:`~.xso.AbstractTextChild.lang` and
    :attr:`~.xso.AbstractTextChild.text`, respectively and support the
    same-named keyword arguments for those attributes at the consturctor.

    For an example see the source of :class:`aioxmpp.Message`.

    .. versionadded:: 0.5
    """

    def __init__(self, xso_type):
        super().__init__()
        self.xso_type = xso_type

    def get_xso_types(self):
        return [self.xso_type]

    def unpack(self, obj):
        return obj.lang, obj.text

    def pack(self, item):
        lang, text = item
        xso = self.xso_type(text=text, lang=lang)
        return xso


class EnumCDataType(AbstractCDataType):
    """
    Use an :class:`enum.Enum` as type for an XSO descriptor.

    :param enum_class: The :class:`~enum.Enum` to use as type.
    :param nested_type: A type which can handle the values of the enumeration
        members.
    :type nested_type: :class:`AbstractCDataType`
    :param allow_coerce: Allow coercion of different types to enumeration
                         values.
    :type allow_coerce: :class:`bool`
    :param deprecate_coerce: Emit :class:`DeprecationWarning` when coercion
                             occurs. Requires (but does not imply)
                             `allow_coerce`.
    :type deprecate_coerce: :class:`int` or :class:`bool`
    :param allow_unknown: If true, unknown values are converted to
                          :class:`Unknown` instances when parsing values from
                          the XML stream.
    :type allow_unknown: :class:`bool`
    :param accept_unknown: If true, :class:`Unknown` instances are passed
                           through :meth:`coerce` and can thus be assigned to
                           descriptors using this type.
    :type accept_unknown: :class:`bool`
    :param pass_unknown: If true, unknown values are accepted unmodified (both
        on the receiving and on the sending side). It is useful for some
        :class:`enum.IntEnum` use cases.
    :type pass_unknown: :class:`bool`

    A descriptor using this type will accept elements from the given
    `enum_class` as values. Upon serialisiation, the :attr:`value` of the
    enumeration element is taken and formatted through the given `nested_type`.

    Normally, :meth:`coerce` will raise :class:`TypeError` for any value which
    is not an instance of `enum_class`. However, if `allow_coerce` is true, the
    value is passed to the `enum_class` constructor and the result is returned;
    the :class:`ValueError` raised from the `enum_class` constructor if an
    invalid value is passed propagates unmodified.

    .. note::

       When using `allow_coerce`, keep in mind that this may have surprising
       effects for users. Coercion means that the value assigned to an
       attribute and the value subsequently read from that attribute may not
       be the same; this may be very surprising to users::

         class E(enum.Enum):
             X = "foo"

         class SomeXSO(xso.XSO):
             attr = xso.Attr("foo", xso.EnumCDataType(E, allow_coerce=True))

         x = SomeXSO()
         x.attr = "foo"
         assert x.attr == "foo"  # assertion fails!

    To allow coercion transitionally while moving from e.g. string-based values
    to a proper enum, `deprecate_coerce` can be used. In that case, a
    :class:`DeprecationWarning` (see :mod:`warnings`) is emitted when coercion
    takes place, to warn users about future removal of the coercion capability.
    If `deprecate_coerce` is an integer, it is used as the stacklevel argument
    for the :func:`warnings.warn` call. If it is :data:`True`, the stacklevel
    is 4, which leads to the warning pointing to a descriptor assignment when
    used with XSO descriptors.

    Handling of :class:`Unknown` values: Using `allow_unknown` and
    `accept_unknown` is advisable to stay compatible with future protocols,
    which is why both are enabled by default. Considering that constructing an
    :class:`Unknown` value needs to be done explicitly in code, it is unlikely
    that a user will *accidentally* assign an unspecified value to a descriptor
    using this type with `accept_unknown`.

    `pass_unknown` requires `allow_unknown` and `accept_unknown`. When set to
    true, values which are not a member of `enum_class` are used without
    modification (but they are validated against the `nested_type`). This
    applies to both the sending and the receiving side. The intended use case
    is with :class:`enum.IntEnum` classes. If a :class:`Unknown` value is
    passed, it is unwrapped and treated as if the original value had been
    passed.

    Example::

      class SomeEnum(enum.Enum):
          X = 1
          Y = 2
          Z = 3

      class SomeXSO(xso.XSO):
          attr = xso.Attr(
              "foo",
              type_=xso.EnumCDataType(
                  SomeEnum,
                  # have to use integer, because the value of e.g. SomeEnum.X
                  # is integer!
                  xso.Integer()
              ),
          )

    .. versionchanged:: 0.10

        Support for `pass_unknown` was added.
    """

    def __init__(self, enum_class, nested_type=String(), *,
                 allow_coerce=False,
                 deprecate_coerce=False,
                 allow_unknown=True,
                 accept_unknown=True,
                 pass_unknown=False):
        if pass_unknown and (not allow_unknown or not accept_unknown):
            raise ValueError(
                "pass_unknown requires allow_unknown and accept_unknown"
            )

        super().__init__()
        self.nested_type = nested_type
        self.enum_class = enum_class
        self.allow_coerce = allow_coerce
        self.deprecate_coerce = deprecate_coerce
        self.accept_unknown = accept_unknown
        self.allow_unknown = allow_unknown
        self.pass_unknown = pass_unknown

    def coerce(self, value):
        if (not self.pass_unknown and self.accept_unknown and
                isinstance(value, Unknown)):
            return value

        if self.allow_coerce:
            if self.deprecate_coerce:
                if isinstance(value, self.enum_class):
                    return value
                stacklevel = (4 if self.deprecate_coerce is True
                              else self.deprecate_coerce)
                warnings.warn(
                    "assignment of non-enum values to this descriptor is"
                    " deprecated",
                    DeprecationWarning,
                    stacklevel=stacklevel,
                )

            value = self.nested_type.coerce(value)
            try:
                return self.enum_class(value)
            except ValueError:
                if self.pass_unknown:
                    return value
                raise

        if isinstance(value, self.enum_class):
            return value

        if self.pass_unknown:
            value = self.nested_type.coerce(value)
            return value

        raise TypeError("not a valid {} value: {!r}".format(
            self.enum_class,
            value,
        ))

    def parse(self, s):
        parsed = self.nested_type.parse(s)
        try:
            return self.enum_class(parsed)
        except ValueError:
            if self.pass_unknown:
                return parsed
            if self.allow_unknown:
                return Unknown(parsed)
            raise

    def format(self, v):
        if self.pass_unknown and not isinstance(v, self.enum_class):
            return self.nested_type.format(v)
        return self.nested_type.format(v.value)


class EnumElementType(AbstractElementType):
    """
    Use an :class:`enum.Enum` as type for an XSO descriptor.

    :param enum_class: The :class:`~enum.Enum` to use as type.
    :param nested_type: Type which describes the value type of the
        `enum_class`.
    :type nested_type: :class:`AbstractElementType`
    :param allow_coerce: Allow coercion of different types to enumeration
                         values.
    :type allow_coerce: :class:`bool`
    :param deprecate_coerce: Emit :class:`DeprecationWarning` when coercion
                             occurs. Requires (but does not imply)
                             `allow_coerce`.
    :type deprecate_coerce: :class:`int` or :class:`bool`
    :param allow_unknown: If true, unknown values are converted to
                          :class:`Unknown` instances when parsing values from
                          the XML stream.
    :type allow_unknown: :class:`bool`
    :param accept_unknown: If true, :class:`Unknown` instances are passed
                           through :meth:`coerce` and can thus be assigned to
                           descriptors using this type.
    :type allow_unknown: :class:`bool`

    A descriptor using this type will accept elements from the given
    `enum_class` as values. Upon serialisiation, the :attr:`value` of the
    enumeration element is taken and packed through the given `nested_type`.

    Normally, :meth:`coerce` will raise :class:`TypeError` for any value which
    is not an instance of `enum_class`. However, if `allow_coerce` is true, the
    value is passed to the `enum_class` constructor and the result is returned;
    the :class:`ValueError` raised from the `enum_class` constructor if an
    invalid value is passed propagates unmodified.

    .. seealso::

        :class:`EnumCDataType`
            for a detailed discussion on the implications of coercion.

    Handling of :class:`Unknown` values: Using `allow_unknown` and
    `accept_unknown` is advisable to stay compatible with future protocols,
    which is why both are enabled by default. Considering that constructing an
    :class:`Unknown` value needs to be done explicitly in code, it is unlikely
    that a user will *accidentally* assign an unspecified value to a descriptor
    using this type with `accept_unknown`.
    """

    def __init__(self, enum_class, nested_type, *,
                 allow_coerce=False,
                 deprecate_coerce=False,
                 allow_unknown=True,
                 accept_unknown=True):
        super().__init__()
        self.nested_type = nested_type
        self.enum_class = enum_class
        self.allow_coerce = allow_coerce
        self.deprecate_coerce = deprecate_coerce
        self.accept_unknown = accept_unknown
        self.allow_unknown = allow_unknown

    def get_xso_types(self):
        return self.nested_type.get_xso_types()

    def coerce(self, value):
        if self.accept_unknown and isinstance(value, Unknown):
            return value
        if self.allow_coerce:
            if self.deprecate_coerce:
                if isinstance(value, self.enum_class):
                    return value
                stacklevel = (4 if self.deprecate_coerce is True
                              else self.deprecate_coerce)
                warnings.warn(
                    "assignment of non-enum values to this descriptor is"
                    " deprecated",
                    DeprecationWarning,
                    stacklevel=stacklevel,
                )
            return self.enum_class(value)
        if isinstance(value, self.enum_class):
            return value
        raise TypeError("not a valid {} value: {!r}".format(
            self.enum_class,
            value,
        ))

    def unpack(self, s):
        parsed = self.nested_type.unpack(s)
        try:
            return self.enum_class(parsed)
        except ValueError:
            if self.allow_unknown:
                return Unknown(parsed)
            raise

    def pack(self, v):
        return self.nested_type.pack(v.value)


class AbstractValidator(metaclass=abc.ABCMeta):
    """
    This is the interface all validators must implement. In addition, a
    validators documentation should clearly state on which types it operates.

    .. automethod:: validate

    .. automethod:: validate_detailed
    """

    def validate(self, value):
        """
        Return :data:`True` if the `value` adheres to the restrictions imposed
        by this validator and :data:`False` otherwise.

        By default, this method calls :meth:`validate_detailed` and returns
        :data:`True` if :meth:`validate_detailed` returned an empty result.
        """
        return not self.validate_detailed(value)

    @abc.abstractmethod
    def validate_detailed(self, value):
        """
        Return an empty list if the `value` adheres to the restrictions imposed
        by this validator.

        If the value does not comply, return a list of
        :class:`~aioxmpp.errors.UserValueError` instances which each represent
        a condition which was violated in a human-readable way.
        """


class RestrictToSet(AbstractValidator):
    """
    Restrict the possible values to the values from `values`. Operates on any
    types.
    """

    def __init__(self, values):
        self.values = frozenset(values)

    def validate_detailed(self, value):
        from ..errors import UserValueError
        if value not in self.values:
            return [
                UserValueError(i18n._("{} is not an allowed value"),
                               value)
            ]
        return []


class Nmtoken(AbstractValidator):
    """
    Restrict the possible strings to the NMTOKEN specification of XML Schema
    Definitions. The validator only works with strings.

    .. warning::

       This validator is probably incorrect. It is a good first line of defense
       to avoid creating obvious incorrect output and should not be used as
       input validator.

       It most likely falsely rejects valid values and may let through invalid
       values.

    """

    VALID_CATS = {
        "Ll", "Lu", "Lo", "Lt", "Nl",  # Name start
        "Mc", "Me", "Mn", "Lm", "Nd",  # Name without name start
    }
    ADDITIONAL = frozenset(":_.-\u06dd\u06de\u06df\u00b7\u0387\u212e")
    UCD = unicodedata.ucd_3_2_0

    @classmethod
    def _validate_chr(cls, c):
        if c in cls.ADDITIONAL:
            return True
        if 0xf900 < ord(c) < 0xfffe:
            return False
        if 0x20dd <= ord(c) <= 0x20e0:
            return False
        if cls.UCD.category(c) not in cls.VALID_CATS:
            return False
        return True

    def validate_detailed(self, value):
        from ..errors import UserValueError
        if not all(map(self._validate_chr, value)):
            return [
                UserValueError(i18n._("{} is not a valid NMTOKEN"),
                               value)
            ]
        return []


class IsInstance(AbstractValidator):
    """
    This validator checks that the value is an instance of any of the classes
    given in `valid_classes`.

    `valid_classes` is *not* copied into the :class:`IsInstance` instance, but
    instead shared; it can be mutated after the construction of
    :class:`IsInstance` to allow addition and removal of classes.
    """

    def __init__(self, valid_classes):
        self.classes = valid_classes

    def validate_detailed(self, v):
        from ..errors import UserValueError
        if not isinstance(v, tuple(self.classes)):
            return [
                UserValueError(
                    i18n._("{} is of incorrect type (must be one of {})"),
                    v,
                    ", ".join(type_.__name__
                              for type_ in self.classes)
                )
            ]
        return []


class NumericRange(AbstractValidator):
    """
    To be used with orderable types, such as :class:`.DateTime` or
    :class:`.Integer`.

    The value is enforced to be within *[min, max]* (this is the interval from
    `min_` to `max_`, including both ends).

    Setting `min_` or `max_` to :data:`None` disables enforcement of that end
    of the interval. A common use is ``NumericRange(min_=1)`` in conjunction
    with :class:`.Integer` to enforce the use of positive integers.

    .. versionadded:: 0.6

    """

    def __init__(self, min_=None, max_=None):
        super().__init__()
        self.min_ = min_
        self.max_ = max_

    def validate_detailed(self, v):
        from ..errors import UserValueError
        if self.min_ is None:
            if self.max_ is None:
                return []
            if not v <= self.max_:
                return [
                    UserValueError(
                        i18n._("{} is too large (max is {})"),
                        v,
                        self.max_
                    )
                ]
        elif self.max_ is None:
            if not self.min_ <= v:
                return [
                    UserValueError(
                        i18n._("{} is too small (min is {})"),
                        v,
                        self.max_
                    )
                ]
        elif not self.min_ <= v <= self.max_:
            return [
                UserValueError(
                    i18n._("{} is out of bounds ({}..{})"),
                    v,
                    self.min_,
                    self.max_
                )
            ]
        return []


_Undefined = object()


def EnumType(enum_class, nested_type=_Undefined, **kwargs):
    """
    Create and return a :class:`EnumCDataType` or :class:`EnumElementType`,
    depending on the type of `nested_type`.

    If `nested_type` is a :class:`AbstractCDataType` or omitted, a
    :class:`EnumCDataType` is constructed. Otherwise, :class:`EnumElementType`
    is used.

    The arguments are forwarded to the respective class’ constructor.

    .. versionadded:: 0.10

    .. deprecated:: 0.10

        This function was introduced to ease the transition in 0.10 from
        a unified :class:`EnumType` to split :class:`EnumCDataType` and
        :class:`EnumElementType`.

        It will be removed in 1.0.
    """

    if nested_type is _Undefined:
        return EnumCDataType(enum_class, **kwargs)
    if isinstance(nested_type, AbstractCDataType):
        return EnumCDataType(enum_class, nested_type, **kwargs)
    else:
        return EnumElementType(enum_class, nested_type, **kwargs)



########################################################################
# File name: quickstart_echo_bot.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import getpass

try:
    import readline  # NOQA
except ImportError:
    pass

import aioxmpp


async def main(local_jid, password):
    client = aioxmpp.PresenceManagedClient(
        local_jid,
        aioxmpp.make_security_layer(password)
    )

    def message_received(msg):
        if not msg.body:
            # do not reflect anything without a body
            return

        # we could also use reply = msg.make_reply() instead
        reply = aioxmpp.Message(
            type_=msg.type_,
            to=msg.from_,
        )

        # make_reply() would not set the body though
        reply.body.update(msg.body)

        client.enqueue(reply)

    message_dispatcher = client.summon(
        aioxmpp.dispatcher.SimpleMessageDispatcher
    )
    message_dispatcher.register_callback(
        aioxmpp.MessageType.CHAT,
        None,
        message_received,
    )

    async with client.connected():
        while True:
            await asyncio.sleep(1)


if __name__ == "__main__":
    local_jid = aioxmpp.JID.fromstr(input("local JID> "))
    password = getpass.getpass()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main(local_jid, password))
    loop.close()



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp
import aioxmpp.xso

from aioxmpp.utils import namespaces

namespaces.xep0191 = "urn:xmpp:blocking"


# this XSO represents a single block list item.
class BlockItem(aioxmpp.xso.XSO):
    # define the tag we are matching for
    # tags consist of an XML namespace URI and an XML element
    TAG = (namespaces.xep0191, "item")

    # bind the ``jid`` python attribute to refer to the ``jid`` XML attribute.
    # in addition, automatic conversion between actual JID objects and XML
    # character data is requested by specifying the `type_` argument as
    # xso.JID() object.
    jid = aioxmpp.xso.Attr(
        "jid",
        type_=aioxmpp.xso.JID()
    )


# we now declare a custom type to convert between JID objects and BlockItem
# instances.
# we can use this custom type together with xso.ChildValueList to access the
# list of <item xmlns="urn:xmpp:blocking" /> elements like a normal python list
# of JIDs.
class BlockItemType(aioxmpp.xso.AbstractElementType):
    # unpack converts from the "raw" XSO to the
    # "rich" python representation, in this case a JID object
    # think of unpack like of a high-level struct.unpack: we convert
    # wire-format (XML trees) to python values
    def unpack(self, item):
        return item.jid

    # pack is the reverse operation of unpack
    def pack(self, jid):
        item = BlockItem()
        item.jid = jid
        return item

    # we have to tell the XSO framework what XSO types are supported by this
    # element type
    def get_xso_types(self):
        return [BlockItem]


# the decorator tells the IQ stanza class that this is a valid payload; that is
# required to be able to *receive* payloads of this type (sending works without
# that decorator, but is not recommended)
@aioxmpp.stanza.IQ.as_payload_class
class BlockList(aioxmpp.xso.XSO):
    TAG = (namespaces.xep0191, "blocklist")

    # this does not get an __init__ method, since the client never
    # creates a BlockList with entries.

    # xso.ChildValueList uses an AbstractElementType (like the one we defined
    # above) to convert between child XSO instances and other python objects.
    # it is accessed like a normal list, but when parsing/serialising, the
    # elements are converted to XML structures using the given type.
    items = aioxmpp.xso.ChildValueList(
        BlockItemType()
    )


@aioxmpp.stanza.IQ.as_payload_class
class BlockCommand(aioxmpp.xso.XSO):
    TAG = (namespaces.xep0191, "block")

    def __init__(self, jids_to_block=None):
        if jids_to_block is not None:
            self.items[:] = jids_to_block

    items = aioxmpp.xso.ChildValueList(
        BlockItemType()
    )


@aioxmpp.stanza.IQ.as_payload_class
class UnblockCommand(aioxmpp.xso.XSO):
    TAG = (namespaces.xep0191, "unblock")

    def __init__(self, jids_to_block=None):
        if jids_to_block is not None:
            self.items[:] = jids_to_block

    items = aioxmpp.xso.ChildValueList(
        BlockItemType()
    )
